"""
Shared content data between decoder and encoder.
"""

from abc import ABC
from collections import defaultdict
from dataclasses import MISSING, dataclass, field, fields
from typing import Generic, TypeAlias, cast

from scfile.enums import FileType
from scfile.structures.scene import ModelFlags, ModelScene
from scfile.structures.texture import CubemapTexture, DefaultTexture, TextureType


NbtValue: TypeAlias = None | int | float | bytes | str | list[int] | list["NbtValue"] | dict[str, "NbtValue"]


@dataclass
class FileContent(ABC):
    """Base dataclass for file content storage."""

    type: FileType = field(default=FileType.NONE)

    def reset(self):
        for f in fields(self):
            if f.default_factory is not MISSING:
                setattr(self, f.name, f.default_factory())

            elif f.default is not MISSING:
                setattr(self, f.name, f.default)

            else:
                setattr(self, f.name, None)


@dataclass
class ModelContent(FileContent):
    """Model content storage."""

    type: FileType = field(default=FileType.MODEL)

    version: float = 0.0
    flags: ModelFlags = field(default_factory=lambda: defaultdict(bool))
    scene: ModelScene = field(default_factory=ModelScene)


@dataclass
class TextureContent(FileContent, Generic[TextureType]):
    """Texture content storage."""

    type: FileType = field(default=FileType.TEXTURE)

    width: int = 0
    height: int = 0
    mipmap_count: int = 0
    format: bytes = field(default_factory=bytes)
    texture: TextureType = field(default_factory=lambda: cast(TextureType, DefaultTexture()))

    @property
    def is_cubemap(self) -> bool:
        return isinstance(self.texture, CubemapTexture)

    @property
    def is_compressed(self) -> bool:
        return self.fourcc in (b"DXT1", b"DXT3", b"DXT5", b"ATI2", b"DX10")

    @property
    def fourcc(self) -> bytes:
        match self.format:
            case b"DXN_XY":
                return b"ATI2"
            case b"RGBA32F":
                return b"DX10"
            case _:
                return self.format


@dataclass
class ImageContent(FileContent):
    """Image content storage."""

    type: FileType = field(default=FileType.IMAGE)

    image: bytes = field(default_factory=bytes)


@dataclass
class TextureArrayContent(FileContent):
    """Texture array storage."""

    type: FileType = field(default=FileType.TEXARR)

    count: int = 0
    textures: list[tuple[str, bytes]] = field(default_factory=list)


@dataclass
class NbtContent(FileContent):
    type: FileType = field(default=FileType.NBT)

    value: NbtValue = None



from scfile.consts import FileSignature
from scfile.core import FileDecoder, ImageContent
from scfile.enums import FileFormat


class MicDecoder(FileDecoder[ImageContent]):
    format = FileFormat.MIC
    signature = FileSignature.MIC

    _content = ImageContent

    def to_png(self):
        from scfile.formats.png.encoder import PngEncoder
        return self.convert_to(PngEncoder)

    def parse(self):
        self.data.image = self.read()



import numpy as np

from scfile.consts import FileSignature, McsaModel
from scfile.core import FileEncoder, ModelContent
from scfile.enums import F, FileFormat

from .io import Ms3dFileIO


VERSION = 4
COMMENTS_VERSION = 1
VERTEX_EXTRA_VERSION = 1

MAX_VERTICES = 0xFFFF
MAX_TRIANGLES = 0xFFFF


class Ms3dEncoder(FileEncoder[ModelContent], Ms3dFileIO):
    format = FileFormat.MS3D
    signature = FileSignature.MS3D

    def prepare(self):
        self.data.scene.ensure_unique_names()

        if self._skeleton_presented:
            self.data.scene.skeleton.convert_to_local()

    def serialize(self):
        self._writeb(F.I32, VERSION)
        self._add_vertices()
        self._add_triangles()
        self._add_groups()
        self._add_materials()
        self._add_bones()
        self._add_comments()
        self._add_links()

    def _add_vertices(self):
        # vertices count
        self._writecount("vertices", self.data.scene.total_vertices, MAX_VERTICES)

        # i8 flags, f32 pos[3], i8 bone id, u8 reference count
        fmt = f"{F.I8}{F.F32 * 3}{F.I8}{F.U8}"

        reference_count = 0xFF  # ? necessary only for optimization, calculation too expensive

        for mesh in self.data.scene.meshes:
            for index, xyz in enumerate(mesh.positions):
                bone_id = mesh.links_ids.astype(F.I8)[index][0] if self._skeleton_presented else McsaModel.ROOT_BONE_ID
                self._writeb(fmt, 0, *xyz, bone_id, reference_count)

    def _add_triangles(self):
        # polygons count
        self._writecount("polygons", self.data.scene.total_polygons, MAX_TRIANGLES)

        # u16 flags, u16 indices[3]
        # f32 normals[3][3], f32 textures u[3], f32 textures v[3]
        # u8 smoothing group, u8 group index
        fmt = f"{F.U16 * 4}{F.F32 * 15}{F.U8 * 2}"

        offset = 0
        for index, mesh in enumerate(self.data.scene.meshes):
            for abc in mesh.polygons:
                normals = [i for vertex in abc for i in mesh.normals[vertex]]
                uv = np.concatenate([mesh.textures[abc][:, 0], mesh.textures[abc][:, 1]], dtype=F.F32)
                indices = (abc + offset).astype(F.U16)

                self._writeb(fmt, 0, *indices, *normals, *uv, 1, index)

            offset += mesh.count.vertices

    def _add_groups(self):
        self._writeb(F.U16, len(self.data.scene.meshes))  # groups count

        offset = 0
        for index, mesh in enumerate(self.data.scene.meshes):
            self._writeb(F.U8, 0)  # flags
            self._writefixedstring(mesh.name)  # group name

            count = mesh.count.polygons
            self._writeb(F.U16, count)  # triangles count
            self._writeb(f"{count}{F.U16}", *np.arange(count, dtype=F.U16) + offset)  # indices
            self._writeb(F.I8, index)  # material index

            offset += count

    def _add_materials(self):
        self._writeb(F.U16, len(self.data.scene.meshes))  # materials count

        # f32 ambient[4], diffuse[4], specular[4], emissive[4] (RGBA)
        # f32 shininess, f32 transparency, i8 mode
        fmt = f"{F.F32 * 18}{F.I8}"

        # rgba templates
        empty = (0.0, 0.0, 0.0, 1.0)
        diffuse = (0.8, 0.8, 0.8, 1.0)

        for mesh in self.data.scene.meshes:
            self._writefixedstring(mesh.material)  # material name
            self._writeb(fmt, *empty, *diffuse, *empty, *empty, 0.0, 1.0, 1)
            self._writenull(size=128)  # texture
            self._writenull(size=128)  # alphamap

    def _add_bones(self):
        # f32 fps, f32 frame, f32 framesCount, u16 bonesCount
        fmt = f"{F.F32 * 3}{F.U16}"
        self._writeb(fmt, 24, 1, 30, self.data.scene.count.bones)

        for bone in self.data.scene.skeleton.bones:
            self._writeb(F.U8, 0)  # flags
            self._writefixedstring(bone.name)  # bone name

            parent = self.data.scene.skeleton.bones[bone.parent_id]
            parent_name = parent.name if bone.parent_id != McsaModel.ROOT_BONE_ID else ""
            self._writefixedstring(parent_name)  # parent name

            # f32 bone rotation[3], f32 bone position[3]
            # u16 keyframes rotations, u16 keyframes transitions
            fmt = f"{F.F32 * 6}{F.U16 * 2}"

            qx, qy, qz, qw = bone.quaternion
            self._writeb(fmt, qx, qy, qz, *bone.position, 0, 0)

    def _add_comments(self):
        self._writeb(F.I32, COMMENTS_VERSION)  # comments version
        fmt = F.U32 * 4  # u32 group, u32 material, u32 joints, u32 model
        self._writeb(fmt, 0, 0, 0, 0)  # comments count

    def _add_links(self):
        self._writeb(F.I32, VERTEX_EXTRA_VERSION)  # vertex extra version

        # i8 ids[3], u8 weights[3]
        fmt = f"{F.I8 * 3}{F.U8 * 3}"

        for mesh in self.data.scene.meshes:
            links_ids = mesh.links_ids.astype(F.I8)
            links_weights = (mesh.links_weights * 255).astype(F.U8)

            for ids, weights in zip(links_ids, links_weights):
                self._writeb(fmt, *ids[:3], *weights[:3])



from scfile.consts import Factor, FileSignature, McsaModel, McsaUnits
from scfile.core import FileDecoder, ModelContent
from scfile.enums import ByteOrder, F, FileFormat
from scfile.structures.animation import AnimationClip
from scfile.structures.flags import Flag
from scfile.structures.mesh import LocalBoneId, ModelMesh, SkeletonBoneId
from scfile.structures.skeleton import SkeletonBone

from .exceptions import McsaBoneLinksError, McsaVersionUnsupported
from .io import McsaFileIO
from .versions import SUPPORTED_VERSIONS, VERSION_FLAGS


class McsaDecoder(FileDecoder[ModelContent], McsaFileIO):
    format = FileFormat.MCSA
    order = ByteOrder.LITTLE
    signature = FileSignature.MCSA

    _content = ModelContent

    def to_obj(self):
        from scfile.formats.obj.encoder import ObjEncoder

        return self.convert_to(ObjEncoder)

    def to_glb(self):
        from scfile.formats.glb.encoder import GlbEncoder

        return self.convert_to(GlbEncoder)

    def to_dae(self):
        from scfile.formats.dae.encoder import DaeEncoder

        return self.convert_to(DaeEncoder)

    def to_ms3d(self):
        from scfile.formats.ms3d.encoder import Ms3dEncoder

        return self.convert_to(Ms3dEncoder)

    def parse(self):
        self._parse_header()
        self._parse_meshes()

        if self.data.flags[Flag.SKELETON] and self.options.parse_skeleton:
            self._parse_skeleton()

            if self.options.parse_animation and not self.is_eof():
                self._parse_animation()

    def _parse_header(self):
        self._parse_version()
        self._parse_flags()
        self._parse_scales()

    def _parse_version(self):
        self.data.version = self._readb(F.F32)

        if self.data.version not in SUPPORTED_VERSIONS:
            raise McsaVersionUnsupported(self.path, self.data.version)

    def _parse_flags(self):
        flags_count = VERSION_FLAGS.get(self.data.version)

        if not flags_count:
            raise McsaVersionUnsupported(self.path, self.data.version)

        for index in range(flags_count):
            self.data.flags[index] = self._readb(F.BOOL)

    def _parse_scales(self):
        self.data.scene.scale.position = self._readb(F.F32)

        if self.data.flags[Flag.UV]:
            self.data.scene.scale.texture = self._readb(F.F32)

        # ! Unknown Scale
        if self.data.flags[Flag.NORMALS] and self.data.version >= 10.0:
            self.data.scene.scale.unknown = self._readb(F.F32)

    def _parse_meshes(self):
        self.data.scene.count.meshes = self._readb(F.I32)

        for _ in range(self.data.scene.count.meshes):
            self._parse_mesh()

    def _skip_vertices(self, mesh: ModelMesh, units: int):
        self.read(mesh.count.vertices * units)

    def _parse_mesh(self):
        mesh = ModelMesh()

        # Name & Material
        mesh.name = self._readutf8()
        mesh.material = self._readutf8()

        # Skeleton bone indexes
        if self.data.flags[Flag.SKELETON]:
            mesh.count.links = self._readb(F.U8)
            mesh.count.bones = self._readb(F.U8)

            # Local bones mapping
            for index in range(mesh.count.bones):
                mesh.bones[LocalBoneId(index)] = SkeletonBoneId(self._readb(F.U8))

        # Geometry counts
        mesh.count.vertices = self._readcount("vertices")
        mesh.count.polygons = self._readcount("polygons")

        # ? Not exported
        if self.data.flags[Flag.UV]:
            self.data.scene.scale.filtering = self._readb(F.F32)

        # Default origins
        if self.data.version >= 10.0:
            mesh.origin.rotation = self._readarray(F.F32, 3)
            mesh.origin.position = self._readarray(F.F32, 3)

        # Default scale
        if self.data.version >= 11.0:
            mesh.origin.scale = self._readb(F.F32)

        # Geometric vertices
        self._parse_positions(mesh)

        # Texture coordinates
        if self.data.flags[Flag.UV]:
            self._parse_textures(mesh)

        # ! Data Unconfirmed
        # ? Not parsed
        if self.data.flags[Flag.UNKNOWN_5]:
            self._skip_vertices(mesh, units=4)

        # Vertex normals
        if self.data.flags[Flag.NORMALS]:
            self._parse_normals(mesh)

        # ! Data Unconfirmed
        # ? Not parsed
        if self.data.flags[Flag.UNKNOWN_4]:
            self._skip_vertices(mesh, units=4)

        # Vertex links
        if self.data.flags[Flag.SKELETON]:
            self._parse_links(mesh)

        # ! Data Unconfirmed
        # ? Not parsed
        if self.data.flags[Flag.UNKNOWN_6]:
            self._skip_vertices(mesh, units=4)

        # Polygon faces
        mesh.polygons = self._readpolygons(mesh.count.polygons)

        self.data.scene.meshes.append(mesh)

    def _parse_positions(self, mesh: ModelMesh):
        mesh.positions = self._readvertex(
            fmt=F.I16,
            factor=Factor.I16,
            units=McsaUnits.POSITIONS,
            count=mesh.count.vertices,
            scale=self.data.scene.scale.position,
        )[:, :3]

    def _parse_textures(self, mesh: ModelMesh):
        mesh.textures = self._readvertex(
            fmt=F.I16,
            factor=Factor.I16,
            units=McsaUnits.TEXTURES,
            count=mesh.count.vertices,
            scale=self.data.scene.scale.texture,
        )

    def _parse_normals(self, mesh: ModelMesh):
        mesh.normals = self._readvertex(
            fmt=F.I8,
            factor=Factor.I8,
            units=McsaUnits.NORMALS,
            count=mesh.count.vertices,
        )[:, :3]

    def _parse_links(self, mesh: ModelMesh):
        match mesh.count.links:
            case 0:
                pass
            case 1 | 2:
                self._parse_packed_links(mesh)
            case 3 | 4:
                self._parse_plain_links(mesh)
            case _:
                raise McsaBoneLinksError(self.path, mesh.count.links)

    def _parse_packed_links(self, mesh: ModelMesh):
        if self.options.parse_skeleton:
            links = self._readpackedlinks(mesh.count.vertices, mesh.bones)
            mesh.links_ids, mesh.links_weights = links

        else:
            self._skip_vertices(mesh, units=4)

    def _parse_plain_links(self, mesh: ModelMesh):
        if self.options.parse_skeleton:
            links = self._readplainlinks(mesh.count.vertices, mesh.bones)
            mesh.links_ids, mesh.links_weights = links

        else:
            self._skip_vertices(mesh, units=8)

    def _parse_skeleton(self):
        self.data.scene.count.bones = self._readb(F.U8)

        for index in range(self.data.scene.count.bones):
            self._parse_bone(index)

    def _parse_bone(self, index: int):
        bone = SkeletonBone()

        bone.id = index
        bone.name = self._readutf8()

        # ? Bone is root if parent_id points to itself
        # ? self-reference would cause invalid recursion
        parent_id = self._readb(F.U8)
        bone.parent_id = parent_id if parent_id != index else McsaModel.ROOT_BONE_ID

        bone.position, bone.rotation = self._readbone()

        self.data.scene.skeleton.bones.append(bone)

    def _parse_animation(self):
        self.data.scene.count.clips = self._readb(F.I32)

        for _ in range(self.data.scene.count.clips):
            self._parse_clip()

    def _parse_clip(self):
        clip = AnimationClip()

        clip.name = self._readutf8()
        clip.frames = self._readb(F.U32)
        clip.rate = self._readb(F.F32)
        clip.transforms = self._readclip(clip.frames, self.data.scene.count.bones)

        self.data.scene.animation.clips.append(clip)



"""
Internal data handler module. Provides dataclass-based structures representation.
"""

from . import animation, flags, mesh, scene, skeleton, texture, vectors


__all__ = (
    "animation",
    "flags",
    "mesh",
    "scene",
    "skeleton",
    "texture",
    "vectors",
)


from typing import Optional
from xml.etree.ElementTree import Element, SubElement

import numpy as np


def create_source(
    parent: Element,
    id: str,
    name: str,
    data: np.ndarray,
    tag: str = "float_array",
    count: Optional[int] = None,
) -> Element:
    count = count or len(data)
    source = SubElement(parent, "source", id=f"{id}-{name}")
    array = SubElement(source, tag, id=f"{id}-{name}-array", count=str(count))
    array.text = " ".join(map(str, data.flatten()))
    return source


def add_accessor(
    source: Element,
    id: str,
    name: str,
    count: int,
    components: list[str],
    datatype: str,
    stride: Optional[int] = None,
) -> None:
    array_id = f"#{id}-{name}-array"
    stride = stride or len(components)

    common = SubElement(source, "technique_common")
    accessor = SubElement(common, "accessor", source=array_id, count=str(count), stride=str(stride))

    for component in components:
        accessor.append(Element("param", name=component, type=datatype))



"""
Base class for all processable files.
"""

from abc import ABC
from typing import Optional

from scfile.enums import ByteOrder, FileFormat

from .context import UserOptions
from .io import StructIO


class BaseFile(ABC):
    """Abstract base class providing interface for encoders/decoders."""

    format: FileFormat
    order: ByteOrder = StructIO.order
    signature: Optional[bytes] = None
    options: UserOptions



from dataclasses import dataclass

from scfile import exceptions
from scfile.consts import McsaModel


class McsaDecodingError(exceptions.FileError, exceptions.DecodingError):
    """Base exception for MCSA model related errors."""

    @property
    def prefix(self):
        return "Model"


@dataclass
class McsaCountsLimit(McsaDecodingError, exceptions.ParsingError):
    """Raised when model exceeds allowed geometry limits (vertices/polygons count)."""

    type: str
    count: int

    def __str__(self) -> str:
        return (
            f"{super().__str__()} has invalid structure - "
            f"{self.count:,} {self.type} "
            f"(max reasonable: {McsaModel.GEOMETRY_LIMIT:,})."
        )


@dataclass
class McsaBoneLinksError(McsaDecodingError, exceptions.ParsingError):
    """Raised when mesh contain unexpected/invalid links count."""

    links_count: int

    def __str__(self):
        return f"{super().__str__()} has unexpected links count: {self.links_count}."


@dataclass
class McsaVersionUnsupported(McsaDecodingError, exceptions.UnsupportedError):
    """Raised when attempting to parse unsupported model version."""

    version: float

    def __str__(self):
        return f"{super().__str__()} has unsupported version: {self.version}."



"""
Data types for CLI wrapper.
"""

import pathlib
from typing import Iterator, Optional, Sequence, Tuple, TypeAlias

import click

from scfile.consts import OutputFormats


PathType = pathlib.Path

FilesPaths: TypeAlias = Sequence[PathType]
FilesIter: TypeAlias = Iterator[Tuple[PathType, PathType]]
OutputDir: TypeAlias = Optional[PathType]

Files = click.Path(path_type=PathType, dir_okay=True, file_okay=True, exists=True, resolve_path=True)
Output = click.Path(path_type=PathType, dir_okay=True, file_okay=False)
Formats = click.Choice(list(OutputFormats.MODELS), case_sensitive=False)



"""
CLI wrapper module. Responsible for implementation of interaction with internal core.
"""

from .commands import scfile
from . import types, utils


__all__ = (
    "scfile",
    "types",
    "utils",
)



import click
from rich import print

from scfile import __version__
from scfile.consts import CLI


def version_to_emoji(version: str) -> str:
    major, minor, patch = map(int, version.split("."))
    index = (major * 1000 + minor * 100 + patch) % len(EMOJIS)
    return EMOJIS[index]


def callback(ctx: click.Context, param: click.Parameter, value: bool):
    if not value:
        return

    print(f"scfile, version {__version__} {version_to_emoji(__version__)}")
    print(CLI.FORMATS)

    ctx.exit()


EMOJIS = [
    "🍓", "🍒", "🍑", "🍋", "🍉",
    "🍍", "🍇", "🍏", "🥝", "🥥",
    "🌽", "🌭", "🍕", "🍔", "🧀",
    "🥨", "🍤", "🍫", "🍦", "🍩",
    "🍪", "🍭", "🍰", "☕️", "🧃"
]



"""
Shared user options between decoder and encoder.
"""

from dataclasses import dataclass
from typing import Optional

from scfile.consts import DefaultModelFormats, Formats


@dataclass
class UserOptions:
    """Shared user options between decoder and encoder."""

    model_formats: Optional[Formats] = None
    parse_skeleton: bool = False
    parse_animation: bool = False
    overwrite: bool = True

    @property
    def default_model_formats(self):
        if self.parse_skeleton:
            return DefaultModelFormats.SKELETON
        return DefaultModelFormats.STANDARD



from scfile.core import FileDecoder, TextureArrayContent
from scfile.enums import ByteOrder, F, FileFormat
from scfile.formats.zip.encoder import TextureArrayEncoder


DELIMITER = ":"
FORMAT = FileFormat.DDS.suffix


class TextureArrayDecoder(FileDecoder[TextureArrayContent]):
    format = FileFormat.TEXARR
    order = ByteOrder.BIG

    _content = TextureArrayContent

    def to_zip(self):
        return self.convert_to(TextureArrayEncoder)

    def parse(self):
        self.data.count = self._readb(F.U32)

        for _ in range(self.data.count):
            self._parse_texture()

    def _parse_texture(self):
        path = self._readutf8().replace(DELIMITER, "/") + FORMAT
        size = self._readb(F.U32)
        texture = self.read(size)

        self.data.textures.append((path, texture))



"""
CLI wrapper small utils.
"""

from pathlib import Path

import click
from rich import print

from scfile.consts import CLI, NBT_FILENAMES, SUPPORTED_SUFFIXES, Formats
from scfile.enums import L

from . import types


def no_args(ctx: click.Context) -> None:
    """Prints help message when no arguments provided."""
    click.echo(ctx.get_help())
    print(
        "",
        f"{L.INVALID} No files paths arguments provided. Showing help.",
        f"{L.HINT} Next time drag and drop files onto scfile.exe OR use terminal for parameters.",
        sep="\n",
    )
    click.pause(CLI.PAUSE)


def check_feature_unsupported(user_formats: Formats, unsupported_formats: Formats, feature: str) -> None:
    """Checks that user formats contain unsupported feature."""
    matching_formats = list(filter(lambda fmt: fmt in unsupported_formats, user_formats))

    if bool(matching_formats):
        suffixes = ", ".join(map(lambda fmt: fmt.suffix, matching_formats))
        print(L.WARN, f"Specified formats [b]({suffixes})[/] doesn't support {feature}.")


def is_supported(path: Path) -> bool:
    """Checks that file is supported (by suffix)."""
    return path.is_file() and (path.suffix in SUPPORTED_SUFFIXES or path.name in NBT_FILENAMES)


def paths_to_files_map(paths: types.FilesPaths) -> types.FilesIter:
    """Maps parent directories to their supported files."""
    for path in paths:
        path = path.resolve()

        if path.is_file():
            if is_supported(path):
                yield path.parent, path

        elif path.is_dir():
            for file in path.rglob("*"):
                if is_supported(file):
                    yield path, file


def output_to_destination(root: Path, source: Path, output: types.OutputDir, relative: bool, parent: bool) -> types.OutputDir:
    """Output path with source relative subdirectory appended if relative flag."""
    if relative and output:
        basedir = root.parent if parent else root
        return output / source.relative_to(basedir).parent
    return output



"""
Dataclasses for skeletal animation (clips/actions).
"""

from dataclasses import dataclass, field

import numpy as np

from .skeleton import ModelSkeleton


@dataclass
class AnimationClip:
    """Single animation clip with timing and transformation data."""

    name: str = "clip"
    frames: int = 0
    rate: float = 0.33
    transforms: np.ndarray = field(default_factory=lambda: np.empty(0, dtype=np.float32))

    @property
    def times(self):
        return np.arange(self.frames, dtype=np.float32) * self.rate


@dataclass
class ModelAnimation:
    """Animation clips container."""

    clips: list[AnimationClip] = field(default_factory=list)

    def convert_to_local(self, skeleton: ModelSkeleton):
        for clip in self.clips:
            for bone in skeleton.bones:
                clip.transforms[:, bone.id, 4:7] += bone.position



from scfile.enums import ByteOrder, F
from scfile.formats.mcsa.decoder import McsaDecoder


class McsbDecoder(McsaDecoder):
    def prepare(self):
        self._skip_hash_prefix()

    def _skip_hash_prefix(self):
        size = self._readb(F.I32, ByteOrder.BIG)
        self.read(size)



import pathlib
from dataclasses import dataclass
from typing import Optional

from scfile.types import PathLike


class ScFileException(Exception):
    """Base exception for scfile library."""

    pass


class DecodingError(ScFileException):
    """Base exception occurring while file decoding."""

    pass


class EncodingError(ScFileException):
    """Base exception occurring while file encoding."""

    pass


class ParsingError(ScFileException):
    """Base exception occurring due to unexpected file structure."""

    pass


class UnsupportedError(ScFileException):
    """Base exception occurring intentionally for unsupported formats."""

    pass


class BaseIOError(ScFileException):
    """Base exception occurring i/o operations."""

    @property
    def prefix(self) -> str:
        return "File"

    def __str__(self):
        return f"{self.prefix}"


@dataclass
class FileError(BaseIOError):
    """Base exception occurring file i/o operations."""

    file: PathLike

    @property
    def path(self) -> pathlib.Path:
        return pathlib.Path(self.file)

    def __str__(self):
        return f"{super().__str__()} '{self.path.as_posix()}'"


@dataclass
class FileNotFound(FileError):
    """Raised when file doesn't exist."""

    def __str__(self):
        return f"{super().__str__()} not found or doesn't exist."


@dataclass
class EmptyFileError(FileError):
    """Raised when file is empty."""

    def __str__(self):
        return f"{super().__str__()} is empty."


@dataclass
class UnsupportedFormatError(FileError):
    """Raised when file format (by suffix) is not supported."""

    def __str__(self):
        return f"{super().__str__()} has unsupported suffix '{self.path.suffix}'."


@dataclass
class InvalidSignatureError(FileError):
    """Raised when file signature doesn't match expected."""

    actual: bytes
    expected: bytes

    def __str__(self):
        return (
            f"{super().__str__()} has invalid signature - "
            f"'{self.actual.hex().upper()}' != '{self.expected.hex().upper()}'. "
            "(file suffix doesn't match file content)."
        )


@dataclass
class InvalidStructureError(FileError):
    """Raised when file structure is invalid."""

    position: Optional[int] = None

    def __str__(self):
        return f"{super().__str__()} parsing failed{f' at position {self.position}' if self.position else ''}."



import json
from copy import deepcopy
from typing import Any, Optional, TypeAlias

import numpy as np

from scfile.consts import FileSignature
from scfile.core import FileEncoder, ModelContent
from scfile.enums import F, FileFormat
from scfile.structures.flags import Flag

from . import base
from .enums import BufferTarget, ComponentType


VERSION = 2

Node: TypeAlias = dict[str, Any]
BufferView: TypeAlias = dict[str, int]
Accessor: TypeAlias = dict[str, str | int]


class GlbEncoder(FileEncoder[ModelContent]):
    format = FileFormat.GLB
    signature = FileSignature.GLTF

    def prepare(self):
        self.data.scene.ensure_unique_names()

        if self._skeleton_presented:
            self.data.scene.skeleton.convert_to_local()
            self.data.scene.skeleton.build_hierarchy()

        if self._animation_presented:
            self.data.scene.animation.convert_to_local(self.data.scene.skeleton)

    def serialize(self):
        self._add_header()
        self._create_gltf()
        self._add_json_chunk()
        self._add_binary_chunk()
        self._update_total_size()

    def _add_header(self):
        self._writeb(F.U32, VERSION)

        # Total Size Placeholder
        self.ctx["TOTAL_SIZE_POS"] = self.tell()
        self._writeb(F.U32, 0)

    def _update_total_size(self):
        self.seek(self.ctx["TOTAL_SIZE_POS"])
        self._writeb(F.U32, len(self.getvalue()))

    def _add_json_chunk(self):
        # Serialize gltf json
        gltf = json.dumps(self.ctx["GLTF"])
        gltf_bytes = gltf.encode()
        json_length = len(gltf_bytes)

        # Validate padding length
        padding_length = (4 - (json_length % 4)) % 4

        # Write json
        self._writeb(F.U32, json_length + padding_length)
        self.write(b"JSON")
        self.write(gltf_bytes)

        # Add padding if necessary
        if padding_length > 0:
            self.write(b"\x20" * padding_length)

    def _create_gltf(self):
        self.ctx["GLTF"] = deepcopy(base.GLTF)
        self.ctx["BUFFER_VIEW_OFFSET"] = 0

        # Create scene
        scene: Node = deepcopy(base.SCENE)
        self.ctx["GLTF"]["scenes"].append(scene)

        # Create skeleton keys
        if self._skeleton_presented:
            self.ctx["GLTF"]["skins"] = []

        if self._animation_presented:
            self.ctx["GLTF"]["animations"] = []

        # Create nodes
        self._create_nodes()
        self._count_nodes()

        # Write length in buffers
        self.ctx["GLTF"]["buffers"].append(deepcopy(base.BUFFER))
        self.ctx["GLTF"]["buffers"][0]["byteLength"] = self.ctx["BUFFER_VIEW_OFFSET"]

    def _create_nodes(self):
        self._create_meshes()

        if self._skeleton_presented:
            self._create_bones()
            self._create_bindmatrix()

        if self._animation_presented:
            self._create_animation()

    def _count_nodes(self):
        nodes = list(range(self.data.scene.count.meshes))

        if self._skeleton_presented:
            nodes += self.ctx["ROOT_INDEXES"]

        self.ctx["GLTF"]["scenes"][0]["nodes"] = nodes

    def _accessor_index(self) -> int:
        return len(self.ctx["GLTF"]["accessors"])

    def _create_meshes(self):
        for index, mesh in enumerate(self.data.scene.meshes):
            primitive: Node = deepcopy(base.PRIMITIVE)

            # XYZ Position
            primitive["attributes"]["POSITION"] = self._accessor_index()
            self._create_bufferview(byte_length=mesh.count.vertices * 3 * 4)
            self._create_accessor(mesh.count.vertices, "VEC3", array=mesh.positions)

            # UV Texture
            if self.data.flags[Flag.UV]:
                primitive["attributes"]["TEXCOORD_0"] = self._accessor_index()
                self._create_bufferview(byte_length=mesh.count.vertices * 2 * 4)
                self._create_accessor(mesh.count.vertices, "VEC2")

            # XYZ Normals
            if self.data.flags[Flag.NORMALS]:
                primitive["attributes"]["NORMAL"] = self._accessor_index()
                self._create_bufferview(byte_length=mesh.count.vertices * 3 * 4)
                self._create_accessor(mesh.count.vertices, "VEC3")

            # Bone Links
            if self._skeleton_presented and mesh.count.links > 0:
                # Joint Indices
                primitive["attributes"]["JOINTS_0"] = self._accessor_index()
                self._create_bufferview(byte_length=mesh.count.vertices * 4 * 1)
                self._create_accessor(mesh.count.vertices, "VEC4", ComponentType.UBYTE)

                # Joint Weights
                primitive["attributes"]["WEIGHTS_0"] = self._accessor_index()
                self._create_bufferview(byte_length=mesh.count.vertices * 4 * 4)
                self._create_accessor(mesh.count.vertices, "VEC4", ComponentType.FLOAT)

            # ABC Polygons
            primitive["indices"] = self._accessor_index()
            self._create_bufferview(byte_length=mesh.count.polygons * 4 * 3, target=BufferTarget.ELEMENT_ARRAY_BUFFER)
            self._create_accessor(mesh.count.polygons * 3, "SCALAR", ComponentType.UINT32)

            # Create nodes
            primitive["material"] = index
            node: Node = {"name": mesh.name, "mesh": index}

            if self._skeleton_presented and mesh.count.links > 0:
                node["skin"] = 0

            # Add to GLTF
            self.ctx["GLTF"]["nodes"].append(node)
            self.ctx["GLTF"]["meshes"].append(dict(name=mesh.name, primitives=[primitive]))
            self.ctx["GLTF"]["materials"].append(dict(name=mesh.material, pbrMetallicRoughness=base.PBR))

    def _create_bones(self):
        self.ctx["BONE_INDEXES"] = []
        self.ctx["ROOT_INDEXES"] = []

        node_index_offset = self.data.scene.count.meshes

        for index, bone in enumerate(self.data.scene.skeleton.bones, start=node_index_offset):
            node: Node = dict(
                name=bone.name,
                translation=bone.position.tolist(),
                rotation=bone.quaternion.tolist(),
            )

            self.ctx["BONE_INDEXES"].append(index)

            if bone.is_root:
                self.ctx["ROOT_INDEXES"].append(index)

            if bone.children:
                node["children"] = [node_index_offset + child.id for child in bone.children]

            # Add to GLTF
            self.ctx["GLTF"]["nodes"].append(node)

    def _create_bindmatrix(self):
        self.ctx["GLTF"]["skins"].append(
            dict(
                name="Armature",
                inverseBindMatrices=self._accessor_index(),
                joints=self.ctx["BONE_INDEXES"],
            )
        )
        self._create_bufferview(byte_length=self.data.scene.count.bones * 16 * 4, target=None)
        self._create_accessor(self.data.scene.count.bones, "MAT4", ComponentType.FLOAT)

    def _create_animation(self):
        for clip in self.data.scene.animation.clips:
            time_idx = self._accessor_index()
            self._create_bufferview(byte_length=clip.frames * 4, target=None)
            self._create_accessor(clip.frames, "SCALAR", ComponentType.FLOAT)

            sampler_idx = 0
            samplers = []
            channels = []

            for node_index in self.ctx["BONE_INDEXES"]:
                translation_idx = self._accessor_index()
                self._create_bufferview(byte_length=clip.frames * 3 * 4, target=None)
                self._create_accessor(clip.frames, "VEC3", ComponentType.FLOAT)

                rotation_idx = self._accessor_index()
                self._create_bufferview(byte_length=clip.frames * 4 * 4, target=None)
                self._create_accessor(clip.frames, "VEC4", ComponentType.FLOAT)

                samplers.extend(
                    [
                        dict(input=time_idx, output=translation_idx, interpolation="LINEAR"),
                        dict(input=time_idx, output=rotation_idx, interpolation="LINEAR"),
                    ]
                )
                channels.extend(
                    [
                        dict(sampler=sampler_idx, target=dict(node=node_index, path="translation")),
                        dict(sampler=sampler_idx + 1, target=dict(node=node_index, path="rotation")),
                    ]
                )
                sampler_idx += 2

            self.ctx["GLTF"]["animations"].append(dict(name=clip.name, samplers=samplers, channels=channels))

    def _create_bufferview(
        self,
        byte_length: int,
        target: Optional[BufferTarget] = BufferTarget.ARRAY_BUFFER,
    ):
        view: BufferView = dict(
            buffer=0,
            byteLength=byte_length,
            byteOffset=self.ctx["BUFFER_VIEW_OFFSET"],
        )

        if target:
            view["target"] = target.value

        self.ctx["GLTF"]["bufferViews"].append(view)
        self.ctx["BUFFER_VIEW_OFFSET"] += byte_length

    def _create_accessor(
        self,
        count: int,
        accessor_type: str,
        component_type: ComponentType = ComponentType.FLOAT,
        array: Optional[np.ndarray] = None,
    ):
        buffer_view_idx = len(self.ctx["GLTF"]["bufferViews"]) - 1
        accessor: Accessor = dict(
            bufferView=buffer_view_idx,
            count=count,
            componentType=component_type.value,
            type=accessor_type,
        )

        if array is not None:
            accessor["min"] = np.min(array, axis=0).tolist()
            accessor["max"] = np.max(array, axis=0).tolist()

        self.ctx["GLTF"]["accessors"].append(accessor)

    def _add_binary_chunk(self):
        self._add_bin_size()
        self.ctx["BIN_START"] = self.tell()

        self._add_meshes()

        if self._skeleton_presented:
            self._add_bindmatrix()

        if self._animation_presented:
            self._add_animation()

        self.ctx["BIN_END"] = self.tell()
        self._update_bin_size()

    def _add_bin_size(self):
        # BIN Size Placeholder
        self.ctx["BIN_SIZE_POS"] = self.tell()
        self._writeb(F.U32, 0)
        self.write(b"BIN\0")

    def _update_bin_size(self):
        size = self.ctx["BIN_END"] - self.ctx["BIN_START"]
        self.seek(self.ctx["BIN_SIZE_POS"])
        self._writeb(F.U32, size)

    def _add_meshes(self):
        for mesh in self.data.scene.meshes:
            # XYZ Position
            self.write(mesh.positions.tobytes())

            # UV Texture
            if self.data.flags[Flag.UV]:
                self.write(mesh.textures.tobytes())

            # XYZ Normals
            if self.data.flags[Flag.NORMALS]:
                self.write(mesh.normals.tobytes())

            # Bone Links
            if self._skeleton_presented:
                # Joint Indices
                self.write(mesh.links_ids.tobytes())

                # Joint Weights
                self.write(mesh.links_weights.tobytes())

            # ABC Polygons
            self.write(mesh.polygons.flatten().tobytes())

    def _add_bindmatrix(self):
        # Skeleton bones bind matrix
        bind_matrix = self.data.scene.skeleton.inverse_bind_matrices(transpose=True).tobytes()
        self.write(bind_matrix)

    def _add_animation(self):
        for clip in self.data.scene.animation.clips:
            self.write(clip.times.tobytes())

            for bone in self.data.scene.skeleton.bones:
                bone_transforms = clip.transforms[:, bone.id, :]
                rotations, translations = np.split(bone_transforms, [4], axis=1)

                self.write(translations.tobytes())
                self.write(rotations.tobytes())



from scfile.consts import FileSignature
from scfile.core import FileEncoder
from scfile.core.context import ImageContent
from scfile.enums import FileFormat


class PngEncoder(FileEncoder[ImageContent]):
    format = FileFormat.PNG
    signature = FileSignature.PNG

    def serialize(self):
        self.write(self.data.image)



import itertools
import xml.etree.ElementTree as etree
from xml.etree.ElementTree import Element, SubElement

import numpy as np

from scfile.core import FileEncoder, ModelContent
from scfile.enums import FileFormat
from scfile.structures.flags import Flag
from scfile.structures.mesh import ModelMesh
from scfile.structures.skeleton import SkeletonBone, create_transform_matrix

from . import utils


VERSION = "1.5.0"

DECLARATION = b'<?xml version="1.0" encoding="utf-8"?>\n'
XMLNS = "http://www.collada.org/2008/03/COLLADASchema"

UP_AXIS = "Y_UP"

DEFAULT_COLOR = "1 1 1 1"


class DaeEncoder(FileEncoder[ModelContent]):
    format = FileFormat.DAE

    def prepare(self):
        self.data.scene.ensure_unique_names()

        if self.data.flags[Flag.UV]:
            self.data.scene.invert_v_textures()

        if self._skeleton_presented:
            self.data.scene.skeleton.convert_to_local()
            self.data.scene.skeleton.build_hierarchy()

    def serialize(self):
        self.ctx["ROOT"] = Element("COLLADA", xmlns=XMLNS, version=VERSION)
        self.write(DECLARATION)

        self._add_asset()
        self._add_effects()
        self._add_materials()
        self._add_geometries()

        if self._skeleton_presented:
            self._add_controllers()

        self._add_scenes()

        # Render XML
        etree.indent(self.ctx["ROOT"])
        self.write(etree.tostring(self.ctx["ROOT"]))

    def _add_asset(self):
        asset = SubElement(self.ctx["ROOT"], "asset")
        SubElement(asset, "unit", name="meter", meter="1")
        SubElement(asset, "up_axis").text = UP_AXIS

    def _add_effects(self):
        library = SubElement(self.ctx["ROOT"], "library_effects")

        for mesh in self.data.scene.meshes:
            effect = SubElement(library, "effect", id=f"{mesh.material}-effect")
            profile = SubElement(effect, "profile_COMMON")
            technique = SubElement(profile, "technique", sid="common")
            phong = SubElement(technique, "phong")
            diffuse = SubElement(phong, "diffuse")
            color = SubElement(diffuse, "color")
            color.text = DEFAULT_COLOR

    def _add_materials(self):
        library = SubElement(self.ctx["ROOT"], "library_materials")

        for mesh in self.data.scene.meshes:
            material = SubElement(library, "material", id=f"{mesh.material}-material", name=mesh.material)
            SubElement(material, "instance_effect", url=f"#{mesh.material}-effect")

    def _add_mesh_sources(self, mesh: ModelMesh, node: Element):
        # XYZ Positions
        pos_source = utils.create_source(node, mesh.name, "positions", mesh.positions)
        utils.add_accessor(pos_source, mesh.name, "positions", len(mesh.positions), ["X", "Y", "Z"], "float")

        # UV Texture
        if self.data.flags[Flag.UV]:
            tex_source = utils.create_source(node, mesh.name, "texture", mesh.textures)
            utils.add_accessor(tex_source, mesh.name, "texture", len(mesh.textures), ["S", "T"], "float")

        # XYZ Normals
        if self.data.flags[Flag.NORMALS]:
            norm_source = utils.create_source(node, mesh.name, "normals", mesh.normals)
            utils.add_accessor(norm_source, mesh.name, "normals", len(mesh.normals), ["X", "Y", "Z"], "float")

    def _add_triangles(self, mesh: ModelMesh, node: Element):
        vertices = SubElement(node, "vertices", id=f"{mesh.name}-vertices")
        SubElement(vertices, "input", semantic="POSITION", source=f"#{mesh.name}-positions")

        triangles = SubElement(node, "triangles", count=str(mesh.count.polygons), material=f"{mesh.material}-material")

        # Inputs
        SubElement(triangles, "input", semantic="VERTEX", source=f"#{mesh.name}-vertices", offset="0")

        if self.data.flags[Flag.UV]:
            SubElement(triangles, "input", semantic="TEXCOORD", source=f"#{mesh.name}-texture", offset="0")

        if self.data.flags[Flag.NORMALS]:
            SubElement(triangles, "input", semantic="NORMAL", source=f"#{mesh.name}-normals", offset="0")

        # ABC Polygons
        p = SubElement(triangles, "p")
        p.text = " ".join(map(str, mesh.polygons.flatten()))

    def _add_geometries(self):
        library = SubElement(self.ctx["ROOT"], "library_geometries")

        for mesh in self.data.scene.meshes:
            geometry = SubElement(library, "geometry", id=mesh.name, name=mesh.name)
            node = SubElement(geometry, "mesh")

            self._add_mesh_sources(mesh, node)
            self._add_triangles(mesh, node)

    def _add_controllers(self):
        library = SubElement(self.ctx["ROOT"], "library_controllers")

        for mesh in self.data.scene.meshes:
            controller = SubElement(library, "controller", id=f"{mesh.name}-skin", name="Armature")
            skin = SubElement(controller, "skin", source=f"#{mesh.name}")
            self._add_controller_sources(mesh, skin)
            self._add_joints_and_weights(mesh, skin)

    def _add_controller_sources(self, mesh: ModelMesh, skin: Element):
        # Add joint names
        joint_data = np.array([bone.name for bone in self.data.scene.skeleton.bones])
        joint_source = utils.create_source(skin, mesh.name, "joints", joint_data, "Name_array")
        utils.add_accessor(joint_source, mesh.name, "joints", len(joint_data), ["JOINT"], "name")

        # Add bind poses
        bind_data = self.data.scene.skeleton.inverse_bind_matrices(transpose=False)
        bind_source = utils.create_source(skin, mesh.name, "bindposes", bind_data, count=len(bind_data) * 16)
        utils.add_accessor(bind_source, mesh.name, "bindposes", len(bind_data), ["TRANSFORM"], "float4x4", 16)

        # Add weights
        weight_source = utils.create_source(skin, mesh.name, "weights", mesh.links_weights.flatten())
        utils.add_accessor(weight_source, mesh.name, "weights", len(mesh.links_weights), ["WEIGHT"], "float")

    def _add_joints_and_weights(self, mesh: ModelMesh, skin: Element):
        # Add joints
        joints = SubElement(skin, "joints")
        SubElement(joints, "input", semantic="JOINT", source=f"#{mesh.name}-joints")
        SubElement(joints, "input", semantic="INV_BIND_MATRIX", source=f"#{mesh.name}-bindposes")

        # Add vertex weights
        weights = SubElement(skin, "vertex_weights", count=str(mesh.count.vertices))
        SubElement(weights, "input", semantic="JOINT", source=f"#{mesh.name}-joints", offset="0")
        SubElement(weights, "input", semantic="WEIGHT", source=f"#{mesh.name}-weights", offset="1")

        # Add indices
        weight_index = itertools.count()
        bone_indices = [f"{bone_id} {next(weight_index)}" for bone_id in mesh.links_ids.flatten()]

        SubElement(weights, "vcount").text = " ".join(["4"] * mesh.count.vertices)
        SubElement(weights, "v").text = " ".join(bone_indices)

    def _add_scenes(self):
        library = SubElement(self.ctx["ROOT"], "library_visual_scenes")
        visual_scene = SubElement(library, "visual_scene", id="scene", name="Scene")

        if self._skeleton_presented:
            visual_scene = self._add_armature(visual_scene)

        self._add_mesh_instances(visual_scene)

        scene = SubElement(self.ctx["ROOT"], "scene")
        SubElement(scene, "instance_visual_scene", url="#scene")

    def _add_armature(self, scene: Element) -> Element:
        node = SubElement(scene, "node", id="armature", name="Armature", type="NODE")
        SubElement(node, "matrix", sid="transform").text = "1 0 0 0 0 1 0 0 0 0 1 0 0 0 0 1"

        for root in self.data.scene.skeleton.roots:
            self._add_bone(node, root)

        return node

    def _add_bone(self, parent: Element, bone: "SkeletonBone"):
        joint = SubElement(parent, "node", id=f"armature-{bone.name}", sid=bone.name, name=bone.name, type="JOINT")

        matrix = create_transform_matrix(bone.position, bone.rotation)
        SubElement(joint, "matrix", sid="transform").text = " ".join(map(str, matrix.flatten()))

        for child in bone.children:
            self._add_bone(joint, child)

    def _add_mesh_instances(self, parent: Element):
        for mesh in self.data.scene.meshes:
            node = SubElement(parent, "node", id=mesh.name, name=mesh.name, type="NODE")

            if self._skeleton_presented:
                bone = self.data.scene.skeleton.roots[0]
                instance = SubElement(node, "instance_controller", url=f"#{mesh.name}-skin")
                SubElement(instance, "skeleton").text = f"#armature-{bone.name}"
            else:
                instance = SubElement(node, "instance_geometry", url=f"#{mesh.name}", name=mesh.name)

            self._add_bind_material(mesh, instance)

    def _add_bind_material(self, mesh: ModelMesh, instance: Element):
        bind = SubElement(instance, "bind_material")
        technique_common = SubElement(bind, "technique_common")
        SubElement(
            technique_common,
            "instance_material",
            symbol=f"{mesh.material}-material",
            target=f"#{mesh.material}-material",
        )



from dataclasses import dataclass

from scfile import exceptions


class OlDecodingError(exceptions.FileError, exceptions.ParsingError):
    """Base exception for OL texture related errors."""

    @property
    def prefix(self):
        return "Texture"


@dataclass
class OlFormatUnsupported(OlDecodingError, exceptions.UnsupportedError):
    """Raised when texture contains unsupported format."""

    format: bytes

    def __str__(self):
        return f"{super().__str__()} has unsupported format: {self.format}."



from typing import NamedTuple, TypeAlias


Template: TypeAlias = str


class Flags(NamedTuple):
    uv: bool
    normals: bool


TEMPLATE: dict[Flags, Template] = {
    Flags(uv=True, normals=True): "f {a}/{a}/{a} {b}/{b}/{b} {c}/{c}/{c}",
    Flags(uv=True, normals=False): "f {a}/{a} {b}/{b} {c}/{c}",
    Flags(uv=False, normals=True): "f {a}//{a} {b}//{b} {c}//{c}",
    Flags(uv=False, normals=False): "f {a} {b} {c}",
}



import zipfile

from scfile.core import FileEncoder, TextureArrayContent
from scfile.enums import FileFormat


class TextureArrayEncoder(FileEncoder[TextureArrayContent]):
    format = FileFormat.ZIP

    def serialize(self):
        with zipfile.ZipFile(self, mode="w", compression=zipfile.ZIP_STORED) as zip:
            for path, data in self.data.textures:
                zip.writestr(path, data)



"""
CLI wrapper implementation.
"""

import traceback
from typing import Optional

import click
from rich import print

from scfile import convert, exceptions
from scfile.consts import CLI, Formats
from scfile.core import UserOptions
from scfile.enums import L

from . import types, utils, version


@click.command(name="scfile", epilog=CLI.EPILOG)
@click.argument("PATHS", type=types.Files, nargs=-1)
@click.option(
    "-O",
    "--output",
    help="Output results directory.",
    type=types.Output,
)
@click.option(
    "-F",
    "--mdlformat",
    help="Preferred format for models.",
    type=types.Formats,
    multiple=True,
)
@click.option(
    "--relative",
    help="Preserve directory structure from source in output.",
    is_flag=True,
)
@click.option(
    "--parent",
    help="Use parent directory as starting point in relative directory.",
    is_flag=True,
)
@click.option(
    "--skeleton",
    help="Parse armature in models.",
    is_flag=True,
)
@click.option(
    "--animation",
    help="Parse builtin clips in models.",
    is_flag=True,
)
@click.option(
    "--unique",
    help="Ensure file saved with unique name, avoiding overwrites.",
    is_flag=True,
)
@click.option(
    "--version",
    help="Show the version and exit.",
    callback=version.callback,
    is_flag=True,
    is_eager=True,
    expose_value=False,
)
@click.pass_context
def scfile(
    ctx: click.Context,
    paths: types.FilesPaths,
    output: types.OutputDir,
    mdlformat: Optional[Formats],
    relative: bool,
    parent: bool,
    skeleton: bool,
    animation: bool,
    unique: bool,
) -> None:
    # In case program executed without arguments
    if not paths:
        utils.no_args(ctx)
        return

    # Normalize options
    model_formats = mdlformat or None
    if parent: relative = True
    if animation: skeleton = True

    # Relative flag is useless without output path
    if relative and not output:
        print(L.WARN, "Flag [b]--relative[/] requires [b]--output[/] option.")

    # Warn if specified formats has unsupported features
    if model_formats:
        if skeleton: utils.check_feature_unsupported(model_formats, CLI.NON_SKELETAL_FORMATS, "skeleton")
        if animation: utils.check_feature_unsupported(model_formats, CLI.NON_ANIMATION_FORMATS, "animation")

    # Prepare options
    options = UserOptions(
        model_formats=model_formats,
        parse_skeleton=skeleton,
        parse_animation=animation,
        overwrite=not unique,
    )

    # Iterate over each directory to their supported files
    for root, source in utils.paths_to_files_map(paths):
        # Get destination path
        dest = utils.output_to_destination(root, source, output, relative, parent)

        # Convert source file
        try:
            convert.auto(source=source, output=dest, options=options)

        except exceptions.InvalidStructureError as err:
            print(L.ERROR, str(err), CLI.EXCEPTION)

        except exceptions.ScFileException as err:
            print(L.ERROR, str(err))

        except Exception as err:
            traceback.print_exception(err)
            print(L.EXCEPTION, f"File '{source.as_posix()}' {err}.", CLI.EXCEPTION)

        else:
            src_path = source.relative_to(root)
            dst_path = dest or source.parent
            print(L.INFO, f"File '{src_path.as_posix()}' converted to '{dst_path.as_posix()}'.")



from enum import StrEnum, auto


class FileFormat(StrEnum):
    """File suffix (without dot)."""

    DAE = auto()
    DDS = auto()
    GLB = auto()
    JSON = auto()
    MCAL = auto()
    MCSA = auto()
    MCSB = auto()
    MCVD = auto()
    MIC = auto()
    MS3D = auto()
    NBT = auto()
    OBJ = auto()
    OL = auto()
    PNG = auto()
    TEXARR = auto()
    ZIP = auto()
    ITEMNAMES = "dat"  # NBT

    @property
    def suffix(self):
        return f".{self.value.lower()}"


class FileType(StrEnum):
    """File content type."""

    NONE = auto()
    MODEL = auto()
    TEXTURE = auto()
    IMAGE = auto()
    TEXARR = auto()
    NBT = auto()


class FileMode(StrEnum):
    """File open mode."""

    RB = READ = auto()
    WB = WRITE = auto()
    AB = APPEND = auto()
    PLUS = "+"


class ByteOrder(StrEnum):
    """File content bytes order."""

    NATIVE = "@"
    STANDARD = "="
    LITTLE = "<"
    BIG = ">"
    NETWORK = "!"


class UnicodeErrors(StrEnum):
    """Unicode errors handling policy."""

    STRICT = auto()
    """Raise UnicodeDecodeError on invalid bytes"""
    IGNORE = auto()
    """Skip invalid bytes silently"""
    REPLACE = auto()
    """Replace invalid bytes with a replacement marker (�)"""

    BACKSLASHREPLACE = BACKSLASH = auto()
    """Replace with backslash-escaped sequences (\\xHH)"""
    NAMEREPLACE = NAME = auto()
    """Replace with \\N{...} escape sequences"""
    XMLCHARREFREPLACE = XML = auto()
    """Replace with XML/HTML numeric entities (&#...;)"""

    SURROGATEESCAPE = SURROGATE = auto()
    """Preserve invalid bytes as surrogate codes"""


class StructFormat(StrEnum):
    """Native C-type struct format codes."""

    BOOL = "?"
    """boolean: `BOOL` `1 byte` [False, True]"""

    I8 = "b"
    """signed char: `BYTE` `1 byte` [-128, 127]"""
    I16 = "h"
    """signed short: `WORD` `2 bytes` [-32768, 32767]"""
    I32 = "i"
    """signed int: `DWORD` `4 bytes` [-2147483648, 2147483647]"""
    I64 = "q"
    """signed long long: `QWORD` `8 bytes` [-9223372036854775808, 9223372036854775807]"""

    U8 = "B"
    """unsigned char: `BYTE` `1 byte` [0, 255]"""
    U16 = "H"
    """unsigned short: `WORD` `2 bytes` [0, 65535]"""
    U32 = "I"
    """unsigned int: `DWORD` `4 bytes` [0, 4294967295]"""
    U64 = "Q"
    """unsigned long long: `QWORD` `8 bytes` [0, 18446744073709551615]"""

    F16 = "e"
    """float: `half-precision` `2 bytes`"""
    F32 = "f"
    """double: `single-precision` `4 bytes`"""
    F64 = "d"
    """double: `double-precision` `8 bytes`"""


F = StructFormat
"""StructFormat Alias."""


class ConsoleLabel(StrEnum):
    """Colored labels for console output."""

    INFO = "[b blue]INFO:[/]"
    HINT = "[b cyan]HINT:[/]"
    WARN = "[b yellow]WARN:[/]"
    ERROR = "[b red]ERROR:[/]"
    INVALID = "[b red]INVALID INPUT:[/]"
    EXCEPTION = "[b red]UNEXPECTED ERROR:[/]"


L = ConsoleLabel
"""ConsoleLabel Alias."""



from .enums import PrimitiveMode


GLTF = {
    "asset": {"version": "2.0", "generator": "onejeuu@scfile"},
    "scene": 0,
    "scenes": [],
    "nodes": [],
    "materials": [],
    "meshes": [],
    "accessors": [],
    "bufferViews": [],
    "buffers": [],
}

SCENE = {"name": "Scene", "nodes": []}

BUFFER = {"byteLength": 0}

PRIMITIVE = {"attributes": {}, "mode": PrimitiveMode.TRIANGLES.value}

# R, G, B, A
COLOR = [0.5, 0.5, 0.5, 1.0]

# pbrMetallicRoughness
PBR = {"baseColorFactor": COLOR}



import gzip

import zstandard as zstd

from scfile.core import FileDecoder, NbtContent
from scfile.core.io import StructFileIO
from scfile.enums import ByteOrder, FileFormat

from .enums import Tag
from .io import NbtBytesIO


class NbtDecoder(FileDecoder[NbtContent], StructFileIO):
    format = FileFormat.NBT
    order = ByteOrder.LITTLE

    _content = NbtContent

    def to_json(self):
        from scfile.formats.json.encoder import JsonEncoder
        return self.convert_to(JsonEncoder)

    def parse(self):
        data = self._decompress()

        stream = NbtBytesIO(data)

        # Read root tag
        tag = stream._read_tag()
        if tag == Tag.END:
            return

        _ = stream._readutf8()  # Skip name
        self.data.value = stream._parse_tag(tag)

    def _decompress(self):
        data = self.read()

        try:
            # Gzip is standard nbt compression
            data = gzip.decompress(data)

        except Exception:
            try:
                # Some synced configs use zstd
                data = zstd.decompress(data)

            except Exception:
                pass

        return data



from typing import Sequence, TypeAlias

from .enums import FileFormat


Formats: TypeAlias = Sequence[FileFormat]


SUPPORTED_FORMATS: set[FileFormat] = {
    FileFormat.MCSA,
    FileFormat.MCSB,
    FileFormat.MCVD,
    FileFormat.MIC,
    FileFormat.OL,
    FileFormat.TEXARR,
    FileFormat.NBT,
    FileFormat.ITEMNAMES,
}
"""Files formats (suffixes without dot) that can be converted."""

SUPPORTED_SUFFIXES: set[str] = set(map(lambda fmt: fmt.suffix, SUPPORTED_FORMATS))
"""Files suffixes that can be converted."""

NBT_FILENAMES: set[str] = {"itemnames.dat", "prefs", "common", "sd0", "sd1", "sd2", "sd3", "sd4"}
"""Files names that can be converted."""


class FileSignature:
    """File magic number for formats (big-endian)."""

    MCSA = b"MCSA"
    MCAL = b"MCAL"
    MIC = b"\x89MIC"
    OL = b"\x0a\x95\x23\xfd"

    DDS = b"DDS "
    PNG = b"\x89PNG"
    GLTF = b"glTF"
    MS3D = b"MS3D000000"


class Factor:
    """Integer max range for integer type."""

    I8 = 0x7F
    U8 = 0xFF
    I16 = 0x7FFF
    U16 = 0xFFFF


class OlString:
    """Xor encoded zero-end string."""

    SIZE = 17
    """16 encoded bytes and last 0x00 byte."""
    XOR = ord("g")
    NULL = ord("G")


class CubemapFaces:
    """Dds cubemap faces."""

    FACES = {"+x", "-x", "+y", "-y", "+z", "-z"}
    COUNT = len(FACES)


class McsaModel:
    """Mcsa model constants."""

    DECIMALS = 6
    ROOT_BONE_ID = -1
    GEOMETRY_LIMIT = 1_000_000
    """Safety limit to prevent memory overflow on corrupted files."""


class McsaUnits:
    """Mcsa structures elements count."""

    POSITIONS = 4
    TEXTURES = 2
    NORMALS = 4
    POLYGONS = 3
    LINKS = 4
    BONES = 6
    FRAMES = 7


class CLI:
    """Command line interface constants."""

    FORMATS = EPILOG = f"Supported Formats: {sorted(SUPPORTED_SUFFIXES)}"
    EXCEPTION = "[b yellow]Input file appears to be corrupted or invalid.[/]"
    PAUSE = "\nPress any key to exit..."

    NON_SKELETAL_FORMATS: Formats = (FileFormat.OBJ,)
    NON_ANIMATION_FORMATS: Formats = (FileFormat.OBJ, FileFormat.MS3D, FileFormat.DAE)


class OutputFormats:
    """Supported output formats for file data types."""

    MODELS: Formats = (FileFormat.OBJ, FileFormat.GLB, FileFormat.DAE, FileFormat.MS3D)
    TEXTURES: Formats = (FileFormat.DDS,)
    IMAGES: Formats = (FileFormat.PNG,)
    TEXARR: Formats = (FileFormat.ZIP,)


class DefaultModelFormats:
    """Default model formats for cases where no preference is specified."""

    STANDARD: Formats = (FileFormat.OBJ,)
    SKELETON: Formats = (FileFormat.GLB,)



"""
Internal factory decorator for convert functions.
"""

from functools import wraps
from typing import Callable, Optional, Type

from scfile.core import FileDecoder, FileEncoder, UserOptions
from scfile.core.types import Content
from scfile.types import PathLike

from .base import convert


def converter(
    decoder: Type[FileDecoder[Content]],
    encoder: Type[FileEncoder[Content]],
) -> Callable:
    """Factory decorator for base convert function with fixed decoder/encoder."""

    def decorator(func: Callable):
        @wraps(func)
        def wrapper(
            source: PathLike,
            output: Optional[PathLike] = None,
            options: Optional[UserOptions] = None,
        ):
            convert(
                decoder=decoder,
                encoder=encoder,
                source=source,
                output=output,
                options=options,
            )

        return wrapper

    return decorator



"""
Handling low-level I/O operations. Foundational layer of core. Includes structed read/write wrappers.
"""

from .base import StructIO
from .streams import StructBytesIO, StructFileIO


__all__ = (
    "StructIO",
    "StructBytesIO",
    "StructFileIO",
)


import lz4.block

from scfile.consts import CubemapFaces
from scfile.core import TextureContent
from scfile.formats.ol.decoder import BaseOlDecoder
from scfile.structures.texture import CubemapTexture


class OlCubemapDecoder(BaseOlDecoder[CubemapTexture]):
    _content = TextureContent

    def prepare(self):
        self.data.texture = CubemapTexture()

    def _parse_sizes(self):
        self.data.texture.uncompressed = self._readsizescubemap(self.data.mipmap_count)
        self.data.texture.compressed = self._readsizescubemap(self.data.mipmap_count)

    def _parse_mipmaps(self):
        for mipmap in range(self.data.mipmap_count):
            for face in range(CubemapFaces.COUNT):
                self.data.texture.faces[face].append(
                    lz4.block.decompress(
                        self.read(self.data.texture.compressed[mipmap][face]),
                        self.data.texture.uncompressed[mipmap][face],
                    )
                )



import warnings

from .texture import TextureType


warnings.warn(
    "Module 'scfile.structures.types' is deprecated and will be removed in the next minor version. "
    "Please import from 'scfile.structures.texture' instead.",
    DeprecationWarning,
    stacklevel=2,
)

__all__ = ["TextureType"]
__deprecated__ = True



"""
Base class for file decoder (parsing).
"""

from abc import ABC, abstractmethod
from typing import Generic, Optional, Type

from scfile import exceptions
from scfile.enums import FileMode
from scfile.types import PathLike

from .base import BaseFile
from .context import UserOptions
from .encoder import FileEncoder
from .io import StructFileIO
from .types import Content


class FileDecoder(BaseFile, StructFileIO, Generic[Content], ABC):
    """Base class for decoding file content into structured data objects."""

    mode: str = FileMode.READ

    _content: type[Content]

    def __init__(self, file: PathLike, options: Optional[UserOptions] = None):
        """Initialize file decoder with source file and options.

        Arguments:
            file: Path to file that will be decoded. Can be any `path-like` object.
            options (optional): User provided options. If None, default `UserOptions` will be used.

        Initialized:
            file (`PathLike`): The input file path.
            options (`UserOptions`): Decoding options (default or user provided).
            data (`Generic[Content]`): Empty instance of content type for storing decoded data.

        Note:
            Actual decoding doesn't happen during initialization.
            Call `decode()` to perform parsing process.
        """

        self.file = file
        self.options: UserOptions = options or UserOptions()
        self.data: Content = self._content()

        super().__init__(file=self.file, mode=self.mode)

    def decode(self, seek: bool = True) -> Content:
        """Decode file: prepare, validate signature, parse. Returns parsed data."""
        self.prepare()
        self.validate_signature()
        self.parse()
        if seek:
            self.seek(0)
        return self.data

    def convert_to(
        self,
        encoder: Type[FileEncoder[Content]],
        options: Optional[UserOptions] = None,
    ) -> FileEncoder[Content]:
        """Decode and convert to encoder. Returns encoder with open buffer (must be closed)."""
        options = options or self.options
        data = self.decode()
        enc = encoder(data, options)
        enc.encode()
        return enc

    def convert(
        self,
        encoder: type[FileEncoder[Content]],
        options: Optional[UserOptions] = None,
    ) -> bytes:
        """Decode, convert to encoder and return bytes. Closes encoder automatically."""
        options = options or self.options
        enc = self.convert_to(encoder, options)
        content = enc.getvalue()
        enc.close()
        return content

    def prepare(self) -> None:
        """Perform file preparation before parsing. *(e.g. skip bytes)*."""
        pass

    @abstractmethod
    def parse(self) -> None:
        """Parse file content into `self.data`."""
        pass

    def validate_signature(self) -> None:
        """Validate file signature. Raises `EmptyFileError` or `InvalidSignatureError` on failure."""
        if self.filesize <= len(self.signature or bytes()):
            raise exceptions.EmptyFileError(self.path)

        if self.signature:
            read = self.read(len(self.signature))

            if read != self.signature:
                raise exceptions.InvalidSignatureError(self.path, read, self.signature)

    def close(self) -> None:
        """Close file buffer. Same as `FileIO.close()`."""
        super().close()



from enum import IntEnum


class Tag(IntEnum):
    END = 0
    BYTE = 1
    SHORT = 2
    INT = 3
    LONG = 4
    FLOAT = 5
    DOUBLE = 6
    BYTE_ARRAY = 7
    STRING = 8
    LIST = 9
    COMPOUND = 10
    INT_ARRAY = 11
    LONG_ARRAY = 12



from dataclasses import dataclass

from scfile import exceptions


class Ms3dEncodingError(exceptions.BaseIOError, exceptions.EncodingError):
    """Base exception for MS3D model related errors."""

    @property
    def prefix(self):
        return "MS3D Model"


@dataclass
class Ms3dCountsLimit(Ms3dEncodingError, exceptions.UnsupportedError):
    """Raised when model exceeds format limitations."""

    type: str
    count: int
    limit: int

    def __str__(self) -> str:
        return f"{super().__str__()} - {self.type} count exceeds limit: {self.count:,} > {self.limit:,}"



"""
Extensions for MCSA file format with custom struct-based I/O methods.
"""

from typing import TypeAlias

import numpy as np

from scfile.consts import Factor, McsaModel, McsaUnits
from scfile.core.io import StructFileIO
from scfile.enums import F
from scfile.structures.mesh import BonesMapping
from scfile.structures.vectors import LinksIds, LinksWeights

from .exceptions import McsaCountsLimit


Links: TypeAlias = tuple[LinksIds, LinksWeights]


class McsaFileIO(StructFileIO):
    def _readcount(self, type: str) -> int:
        count = self._readb(F.U32)

        # ? Prevent memory overflow
        if count > McsaModel.GEOMETRY_LIMIT:
            raise McsaCountsLimit(self.path, type, count)

        return count

    def _readvertex(self, fmt: str, factor: float, units: int, count: int, scale: float = 1.0):
        # Read array
        data = self._readarray(fmt, count * units)

        # Scale values to floats
        data = data.astype(F.F32) * np.float32(scale / factor)

        # Reshape to vertex[attribute[units]]
        # attribute = position[3] / normal[3] / uv[2]
        return data.reshape(-1, units)

    def _readpolygons(self, count: int):
        units = McsaUnits.POLYGONS

        # ? Validate that indexes fits into U16 range, otherwise use U32.
        indexes = count * units
        fmt = F.U16 if indexes <= Factor.U16 else F.U32

        # Read array
        data = self._readarray(fmt, count * units)

        # Reshape to face[indices[3]]
        return data.astype(F.U32).reshape(-1, units)

    def _readbone(self):
        units = McsaUnits.BONES

        # Read array
        data = self._readarray(F.F32, units)

        # Reshape to bone[position[3], rotation[3]]
        return data.astype(F.F32).reshape(2, 3)

    def _readclip(self, times_count: int, bones_count: int):
        units = McsaUnits.FRAMES

        # Read array
        data = self._readarray(F.I16, times_count * bones_count * units)

        # Scale values to floats
        data = data.astype(F.F32) * np.float32(1.0 / Factor.I16)

        # Reshape to clip[frames][bones][transforms[7]]
        # transforms = [rotation[4], translation[3]]
        return data.reshape(times_count, bones_count, units)

    def _readpackedlinks(self, count: int, bones: BonesMapping) -> Links:
        units = McsaUnits.LINKS

        # Read array
        data = self._readarray(F.U8, count * units)

        # Reshape to vertex[skin[2][2]]
        # skin = [bone_ids[2], weights[2]]
        data = data.reshape(-1, 2, 2)

        # Unpack and pad values
        ids, weights = _padded(data[:, 0, :]), _padded(data[:, 1, :])

        return _links(ids.flatten(), weights.flatten(), bones)

    def _readplainlinks(self, count: int, bones: BonesMapping) -> Links:
        units = McsaUnits.LINKS

        # Read arrays: bone_ids[vertex][units], weights[vertex][units]
        ids = self._readarray(F.U8, count * units)
        weights = self._readarray(F.U8, count * units)

        return _links(ids, weights, bones)


def _padded(arr: np.ndarray) -> np.ndarray:
    width = ((0, 0), (0, max(0, 4 - arr.shape[-1])))
    return np.pad(arr, width, mode="constant")


def _apply_bones_mapping(ids: np.ndarray, bones: BonesMapping) -> LinksIds:
    max_id = max(bones.keys())
    lookup = np.zeros(max_id + 1, dtype=F.U8)

    for k, v in bones.items():
        lookup[k] = v

    mask = np.clip(ids, 0, max_id)
    return lookup[mask]


def _links(ids: np.ndarray, weights: np.ndarray, bones: BonesMapping) -> Links:
    # Convert local bone ids to skeleton bone ids
    ids = _apply_bones_mapping(ids, bones)

    # Clean empty ids
    ids[weights == 0.0] = 0

    # Scale, Round, Normalize
    weights = weights.astype(F.F32) * np.float32(1.0 / Factor.U8)

    # Reshape to vertex[bone_ids[units]], vertex[weights[units]]
    return (ids.astype(F.U8).reshape(-1, 4), weights.astype(F.F32).reshape(-1, 4))



import json

from scfile.core import FileEncoder
from scfile.core.context import NbtContent
from scfile.enums import FileFormat


class JsonEncoder(FileEncoder[NbtContent]):
    format = FileFormat.JSON

    def serialize(self):
        data = json.dumps(self.data.value, default=str, ensure_ascii=False, indent=2)
        data = data.encode()
        self.write(data)



"""
Core module. It defines base classes and their interaction interface.
"""

from . import base, decoder, encoder, types, context, io
from .base import BaseFile
from .decoder import FileDecoder
from .encoder import FileEncoder
from .context.options import UserOptions
from .context.content import FileContent, ModelContent, TextureContent, ImageContent, TextureArrayContent, NbtContent


__all__ = (
    "base",
    "decoder",
    "encoder",
    "types",
    "context",
    "io",
    "BaseFile",
    "FileDecoder",
    "FileEncoder",
    "UserOptions",
    "FileContent",
    "ModelContent",
    "TextureContent",
    "ImageContent",
    "TextureArrayContent",
    "NbtContent",
)



"""
Shared context between decoder and encoder.
"""

from .content import FileContent, ImageContent, ModelContent, TextureContent, TextureArrayContent, NbtContent
from .options import UserOptions


__all__ = (
    "FileContent",
    "ImageContent",
    "ModelContent",
    "TextureContent",
    "TextureArrayContent",
    "NbtContent",
    "UserOptions",
)



"""
Extensions for MS3D file format with custom struct-based I/O methods.
"""

from scfile.core.io import StructBytesIO
from scfile.enums import F

from .exceptions import Ms3dCountsLimit


class Ms3dFileIO(StructBytesIO):
    def _writecount(self, type: str, count: int, limit: int) -> None:
        if count > limit:
            raise Ms3dCountsLimit(type, count, limit)
        self._writeb(F.U16, count)

    def _writefixedstring(self, text: str) -> None:
        self.write(text.encode("utf-8").ljust(32, b"\x00"))



"""
Conversion by input path based on file suffix.
"""

import pathlib
from collections import deque
from typing import Callable, Optional, TypeAlias

import lz4.block

from scfile import exceptions
from scfile.consts import NBT_FILENAMES
from scfile.core import UserOptions
from scfile.enums import FileFormat
from scfile.types import OutputDir, PathLike

from . import formats, legacy


ConvertersMap: TypeAlias = dict[FileFormat, Callable]


MCSB: ConvertersMap = {
    FileFormat.OBJ: formats.mcsb_to_obj,
    FileFormat.GLB: formats.mcsb_to_glb,
    FileFormat.DAE: formats.mcsb_to_dae,
    FileFormat.MS3D: formats.mcsb_to_ms3d,
}

MCSA: ConvertersMap = {
    FileFormat.OBJ: legacy.mcsa_to_obj,
    FileFormat.GLB: legacy.mcsa_to_glb,
    FileFormat.DAE: legacy.mcsa_to_dae,
    FileFormat.MS3D: legacy.mcsa_to_ms3d,
}


def auto(
    source: PathLike,
    output: OutputDir = None,
    options: Optional[UserOptions] = None,
) -> None:
    """
    Automatically convert file between formats based on file suffix.

    Arguments:
        source: Path to input file.
        output (optional): Path to output directory. Defaults: `Same directory as source`.
        options (optional): User settings. Default: `None`.

    Example:
        `auto("model.mcsb", "path/to/output", UserOptions(parse_skeleton=True))`
    """

    src_path = pathlib.Path(source)
    src_format = src_path.suffix.lstrip(".")

    options = options or UserOptions()
    model_formats = options.model_formats or options.default_model_formats

    # Detect format by file suffix
    match src_format:
        case FileFormat.MCSB:
            # Convert MCSB to all requested formats.
            deque(map(lambda fmt: MCSB[fmt](source, output, options), model_formats), maxlen=0)

        case FileFormat.MCSA | FileFormat.MCVD:
            # Convert MCSA to all requested formats.
            deque(map(lambda fmt: MCSA[fmt](source, output, options), model_formats), maxlen=0)

        case FileFormat.OL:
            # Try standard texture first.
            try:
                formats.ol_to_dds(source, output, options)

            except lz4.block.LZ4BlockError:
                # Fallback to cubemap on failure.
                try:
                    formats.ol_cubemap_to_dds(source, output, options)

                except lz4.block.LZ4BlockError as err:
                    raise exceptions.InvalidStructureError(source) from err

        case FileFormat.MIC:
            formats.mic_to_png(source, output, options)

        case FileFormat.TEXARR:
            formats.texarr_to_zip(source, output, options)

        case FileFormat.NBT:
            formats.nbt_to_json(source, output, options)

        case _:
            # Detect nbt by file name
            if src_path.name in NBT_FILENAMES:
                formats.nbt_to_json(source, output, options)
                return

            raise exceptions.UnsupportedFormatError(src_path)



from scfile.consts import FileSignature
from scfile.core import FileEncoder, TextureContent
from scfile.enums import F, FileFormat
from scfile.formats.dds.enums import DXGIDimension, DXGIFormat
from scfile.structures.texture import TextureType

from .header import DDS
from .mask import BGRA8, RGBA8


class DdsEncoder(FileEncoder[TextureContent[TextureType]]):
    format = FileFormat.DDS
    signature = FileSignature.DDS

    def serialize(self):
        self._add_header()
        self._add_pixelformat()
        self._add_caps()

        if self.data.fourcc == b"DX10":
            self._add_dxgi()

        self.write(self.data.texture.image)

    def _add_header(self):
        self._writeb(
            f"{7}{F.U32}",
            DDS.HEADER.SIZE,  # dwSize
            self._flags,  # dwFlags
            self.data.height,  # dwHeight
            self.data.width,  # dwWidth
            self._pitch_or_linear_size,  # dwPitchOrLinearSize
            0,  # dwDepth
            self.data.mipmap_count,  # dwMipMapCount
        )
        self._writenull(size=4 * 11)  # dwReserved1[11]

    def _add_pixelformat(self):
        self._writeb(F.U32, DDS.PF.SIZE)  # dwSize

        if self.data.is_compressed:
            self._add_pf_fourcc()
        else:
            self._add_pf_rgb()

    def _add_pf_fourcc(self):
        self._writeb(F.U32, DDS.PF.FLAG.FOURCC)  # dwFlags
        self.write(self.data.fourcc)  # dwFourCC
        self._writenull(size=4 * 5)  # dwRGBBitCount, RGBA bit masks (unused)

    def _add_pf_rgb(self):
        self._writeb(F.U32, DDS.PF.RGB)  # dwFlags
        self._writenull(size=4)  # dwFourCC (unused)
        self._writeb(F.U32, DDS.PF.BIT_COUNT)  # dwRGBBitCount

        bitmask = BGRA8 if self.data.fourcc == b"BGRA8" else RGBA8  # not best realization...
        self._writeb(F.U32 * 4, *bitmask)  # RGBA bit masks

    def _add_caps(self):
        self._writeb(
            f"{2}{F.U32}",
            DDS.CAPS1,  # dwCaps1
            self._caps2,  # dwCaps2
        )
        self._writenull(size=4 * 3)  # dwCaps3, dwCaps4, Reserved

    def _add_dxgi(self):
        self._writeb(
            f"{5}{F.U32}",
            DXGIFormat.FLOAT_R32G32B32A32,  # dxgiFormat
            DXGIDimension.TEXTURE2D,  # resourceDimension
            0,  # miscFlag
            1,  # arraySize
            0,  # miscFlags2
        )

    @property
    def _flags(self) -> int:
        if self.data.is_compressed:
            return DDS.HEADER.FLAGS | DDS.HEADER.FLAG.LINEARSIZE
        return DDS.HEADER.FLAGS | DDS.HEADER.FLAG.PITCH

    @property
    def _pitch(self) -> int:
        bytes_per_pixel = 4
        aligned_width = (self.data.width * bytes_per_pixel + 3) & ~3
        return aligned_width

    @property
    def _pitch_or_linear_size(self) -> int:
        if self.data.is_compressed:
            return self.data.texture.linear_size
        return self._pitch

    @property
    def _caps2(self) -> int:
        if self.data.is_cubemap:
            return DDS.CUBEMAPS
        return 0



import numpy as np

from scfile.core import FileEncoder, ModelContent
from scfile.enums import FileFormat
from scfile.structures.flags import Flag
from scfile.structures.mesh import ModelMesh

from . import faces


class ObjEncoder(FileEncoder[ModelContent]):
    format = FileFormat.OBJ

    def prepare(self):
        self.data.scene.ensure_unique_names()

        if self.data.flags[Flag.UV]:
            self.data.scene.flip_v_textures()

    def serialize(self):
        self._add_meshes()

    def _add_meshes(self):
        offset = 1

        for mesh in self.data.scene.meshes:
            self._writeutf8(f"o {mesh.name}\n")
            self._writeutf8(f"usemtl {mesh.material}\n")

            self._add_geometric_vertices(mesh)

            if self.data.flags[Flag.UV]:
                self._add_texture_coordinates(mesh)

            if self.data.flags[Flag.NORMALS]:
                self._add_vertex_normals(mesh)

            self._writeutf8(f"g {mesh.name}\n")
            self._add_polygonal_faces(mesh, offset)

            offset += mesh.count.vertices

    def _vectorize(self, template: bytes, data: np.ndarray, count: int):
        return (template * count) % tuple(data.flatten().tolist())

    def _add_geometric_vertices(self, mesh: ModelMesh):
        template = b"v %.6f %.6f %.6f\n"
        self.write(self._vectorize(template, mesh.positions, mesh.count.vertices))
        self.write(b"\n")

    def _add_texture_coordinates(self, mesh: ModelMesh):
        template = b"vt %.6f %.6f\n"
        self.write(self._vectorize(template, mesh.textures, mesh.count.vertices))
        self.write(b"\n")

    def _add_vertex_normals(self, mesh: ModelMesh):
        template = b"vn %.6f %.6f %.6f\n"
        self.write(self._vectorize(template, mesh.normals, mesh.count.vertices))
        self.write(b"\n")

    def _add_polygonal_faces(self, mesh: ModelMesh, offset: int):
        flags = faces.Flags(uv=self.data.flags[Flag.UV], normals=self.data.flags[Flag.NORMALS])
        template = faces.TEMPLATE[flags]

        polygons = mesh.polygons + offset
        self._writeutf8("\n".join([template.format(a=a, b=b, c=c) for a, b, c in polygons.tolist()]))
        self.write(b"\n\n")



"""
Enum for 3D scene flags.
"""

from enum import IntEnum, auto


class Flag(IntEnum):
    SKELETON = 0
    UV = auto()
    NORMALS = auto()
    UNKNOWN_4 = auto()
    UNKNOWN_5 = auto()
    UNKNOWN_6 = auto()


def to_named_dict(flags: dict[int, bool]):
    return {Flag(key).name: value for key, value in flags.items()}



from .enums import CubemapFlag, HeaderFlag, PixelFormatFlag


class Header:
    SIZE = 124
    FLAG = HeaderFlag
    FLAGS = FLAG.CAPS | FLAG.HEIGHT | FLAG.WIDTH | FLAG.PIXELFORMAT | FLAG.MIPMAPCOUNT


class PixelFormat:
    SIZE = 32
    BIT_COUNT = 32
    FLAG = PixelFormatFlag
    RGB = FLAG.RGB | FLAG.ALPHAPIXELS


class DDS:
    HEADER = Header
    PF = PixelFormat
    CF = CubemapFlag
    CUBEMAP = 0x200
    CUBEMAPS = CUBEMAP | CF.POSX | CF.NEGX | CF.POSY | CF.NEGY | CF.POSZ | CF.NEGZ
    COMPLEX = 0x8
    TEXTURE = 0x1000
    MIPMAP = 0x400000
    CAPS1 = TEXTURE | COMPLEX | MIPMAP



import sys

import click
from rich import print

from scfile.cli import scfile
from scfile.enums import L


def main():
    """Program entrypoint."""

    try:
        scfile(standalone_mode=False)

    except click.ClickException as err:
        print(L.INVALID, str(err))

    except (KeyboardInterrupt, click.exceptions.Abort):
        print("[yellow]Operation aborted.[/]")

    sys.exit()


if __name__ == "__main__":
    main()



"""
Extends standard IO streams with struct-based I/O support.
"""

import io
import struct
from pathlib import Path

from scfile import exceptions

from .base import StructIO


class StructBytesIO(io.BytesIO, StructIO):
    """In-memory bytes buffer with structured binary operations."""

    pass


class StructFileIO(io.FileIO, StructIO):
    """File-based I/O with structured binary operations."""

    @property
    def path(self) -> Path:
        return Path(str(self.name))

    @property
    def filesize(self) -> int:
        return self.path.stat().st_size

    def is_eof(self) -> bool:
        return self.filesize <= self.tell()

    def _unpack(self, fmt: str):
        try:
            return super()._unpack(fmt)

        except struct.error as err:
            raise exceptions.InvalidStructureError(self.path, position=self.tell()) from err



"""
Utility module. Handles conversion logic between formats.
"""

from . import base, legacy, detect, formats
from .detect import auto
from .formats import (
    mcsb_to_dae,
    mcsb_to_glb,
    mcsb_to_ms3d,
    mcsb_to_obj,
    ol_to_dds,
    ol_cubemap_to_dds,
    mic_to_png,
    texarr_to_zip,
)


__all__ = (
    "base",
    "legacy",
    "detect",
    "formats",
    "auto",
    "mcsb_to_dae",
    "mcsb_to_glb",
    "mcsb_to_ms3d",
    "mcsb_to_obj",
    "ol_to_dds",
    "ol_cubemap_to_dds",
    "mic_to_png",
    "texarr_to_zip",
)



from enum import IntEnum


class ComponentType(IntEnum):
    UBYTE = 5121
    FLOAT = 5126
    UINT16 = 5123
    UINT32 = 5125


class BufferTarget(IntEnum):
    ARRAY_BUFFER = 34962
    ELEMENT_ARRAY_BUFFER = 34963


class PrimitiveMode(IntEnum):
    TRIANGLES = 4



"""
Basic implementation of converting one format to another.
"""

from pathlib import Path
from typing import Optional, Type

from scfile import exceptions
from scfile.core import FileDecoder, FileEncoder, UserOptions
from scfile.core.types import Content
from scfile.types import OutputDir, PathLike


def convert(
    decoder: Type[FileDecoder[Content]],
    encoder: Type[FileEncoder[Content]],
    source: PathLike,
    output: OutputDir = None,
    options: Optional[UserOptions] = None,
) -> None:
    """
    Converts file between formats with basic validations.

    Arguments:
        decoder: Input file decoder class.
        encoder: Output file encoder class.
        source: Path to input file.
        output (optional): Path to output directory. Defaults: `Same directory as source`.
        options (optional): User settings. Default: `None`.

    Example:
        `convert(McsaDecoder, ObjEncoder, "model.mcsb", "path/to/output")`
    """

    src_path = Path(source)
    out_dir = Path(output or src_path.parent)
    options = options or UserOptions()

    if not src_path.exists() or not src_path.is_file():
        raise exceptions.FileNotFound(src_path)

    if not out_dir.exists():
        out_dir.mkdir(exist_ok=True, parents=True)

    with decoder(file=src_path, options=options) as src:
        with src.convert_to(encoder=encoder) as out:
            output = out_dir / f"{src_path.stem}{out.suffix}"

            if not options.overwrite:
                output = ensure_unique_path(path=output, suffix=out.suffix)

            out.save(path=output)


def ensure_unique_path(path: Path, suffix: str) -> Path:
    """Generates unique file path by appending counter if path exists."""

    filename = path.stem
    counter = 1

    while path.exists():
        path = path.parent / Path(f"{filename} ({counter}){suffix}")
        counter += 1

    return path



from abc import ABC, abstractmethod
from typing import Generic

import lz4.block

from scfile.consts import FileSignature
from scfile.core import FileDecoder, TextureContent
from scfile.enums import ByteOrder, F, FileFormat
from scfile.structures.texture import DefaultTexture, TextureType

from .exceptions import OlFormatUnsupported
from .formats import SUPPORTED_FORMATS
from .io import OlFileIO


# mro nightmare
class BaseOlDecoder(FileDecoder[TextureContent[TextureType]], OlFileIO, Generic[TextureType], ABC):
    format = FileFormat.OL
    order = ByteOrder.BIG
    signature = FileSignature.OL

    def to_dds(self):
        from scfile.formats.dds.encoder import DdsEncoder
        return self.convert_to(DdsEncoder)

    def parse(self):
        self._parse_header()
        self._parse_format()
        self._parse_sizes()
        self._parse_image()

    def _parse_header(self):
        self.data.width = self._readb(F.U32)
        self.data.height = self._readb(F.U32)
        self.data.mipmap_count = self._readb(F.U32)

    def _parse_format(self):
        self.data.format = self._readformat()

        if self.data.format not in SUPPORTED_FORMATS:
            raise OlFormatUnsupported(self.path, self.data.format)

    def _parse_image(self):
        self.texture_id = self._reads()
        self._parse_mipmaps()

    @abstractmethod
    def _parse_sizes(self):
        pass

    @abstractmethod
    def _parse_mipmaps(self):
        pass


class OlDecoder(BaseOlDecoder[DefaultTexture]):
    _content = TextureContent

    def _parse_sizes(self):
        self.data.texture.uncompressed = self._readsizes(self.data.mipmap_count)
        self.data.texture.compressed = self._readsizes(self.data.mipmap_count)

    def _parse_mipmaps(self):
        for mipmap in range(self.data.mipmap_count):
            self.data.texture.mipmaps.append(
                lz4.block.decompress(
                    self.read(self.data.texture.compressed[mipmap]),
                    self.data.texture.uncompressed[mipmap],
                )
            )



"""
Extends IOBase to support struct-based I/O.
"""

import io
import struct
from typing import Any, Optional

import numpy as np

from scfile.enums import ByteOrder, F, UnicodeErrors


class StructIO(io.IOBase):
    """Low-level binary I/O with structured packing/unpacking operations."""

    order: ByteOrder = ByteOrder.LITTLE
    unicode_errors: str = UnicodeErrors.REPLACE

    def _pack(self, fmt: str, *values: Any) -> bytes:
        return struct.pack(str(fmt), *values)

    def _unpack(self, fmt: str) -> tuple[Any, ...]:
        size = struct.calcsize(str(fmt))
        data = self.read(size)
        return struct.unpack(fmt, data)

    def _readarray(self, dtype: str, count: int, order: Optional[ByteOrder] = None):
        order = order or self.order
        datatype = np.dtype(f"{order}{dtype}")
        datasize = count * datatype.itemsize
        return np.frombuffer(self.read(datasize), dtype=datatype, count=count)

    def _readb(self, fmt: str, order: Optional[ByteOrder] = None) -> Any:
        order = order or self.order
        return self._unpack(f"{order}{fmt}")[0]

    def _reads(self, prefix: str = F.U16, order: Optional[ByteOrder] = None) -> bytes:
        order = order or self.order
        size = self._readb(prefix, order)
        return self._unpack(f"{size}s")[0]

    def _readutf8(self) -> str:
        return self._reads().decode("utf-8", errors=self.unicode_errors)

    def _writeb(self, fmt: str, *values: Any, order: Optional[ByteOrder] = None) -> None:
        order = order or self.order
        data = self._pack(f"{order}{fmt}", *values)
        self.write(data)

    def _writenull(self, size: int = 4) -> None:
        self.write(bytes(size))

    def _writeutf8(self, string: str) -> None:
        self.write(string.encode("utf-8", errors=self.unicode_errors))



"""
Collection of submodules that implement specific file format decoder/encoder.
"""

from . import dae, dds, glb, hdri, mcsa, mcsb, mic, ms3d, obj, ol, png, texarr, zip, nbt, json


__all__ = (
    "dae",
    "dds",
    "glb",
    "hdri",
    "mcsa",
    "mcsb",
    "mic",
    "ms3d",
    "obj",
    "ol",
    "png",
    "texarr",
    "zip",
    "nbt",
    "json",
)



"""
Dataclasses for 3D model geometry.
"""

from dataclasses import dataclass, field
from typing import NewType, TypeAlias

import numpy as np

from .vectors import LinksIds, LinksWeights, Polygons, Vector2D, Vector3D


LocalBoneId = NewType("LocalBoneId", int)  # Bone id within local mesh scope
SkeletonBoneId = NewType("SkeletonBoneId", int)  # Bone id within skeleton hierarchy

BonesMapping: TypeAlias = dict[LocalBoneId, SkeletonBoneId]  # Mapping between local and skeleton ids


@dataclass
class MeshCounts:
    """Quantifying of mesh elements."""

    vertices: int = 0
    polygons: int = 0
    links: int = 0
    bones: int = 0


@dataclass
class MeshOrigin:
    position: Vector3D = field(default_factory=lambda: np.empty(3, dtype=np.float32))
    rotation: Vector3D = field(default_factory=lambda: np.empty(3, dtype=np.float32))
    scale: float = 0.0


@dataclass
class ModelMesh:
    """3D mesh geometry container."""

    name: str = "name"
    material: str = "material"

    count: MeshCounts = field(default_factory=MeshCounts)
    origin: MeshOrigin = field(default_factory=MeshOrigin)

    bones: BonesMapping = field(default_factory=dict)

    positions: Vector3D = field(default_factory=lambda: np.empty((0, 3), dtype=np.float32))
    textures: Vector2D = field(default_factory=lambda: np.empty((0, 2), dtype=np.float32))
    normals: Vector3D = field(default_factory=lambda: np.empty((0, 3), dtype=np.float32))

    links_ids: LinksIds = field(default_factory=lambda: np.empty((0, 4), dtype=np.uint8))
    links_weights: LinksWeights = field(default_factory=lambda: np.empty((0, 4), dtype=np.float32))

    polygons: Polygons = field(default_factory=lambda: np.empty((0, 3), dtype=np.uint32))



from typing import TypeVar

from scfile.core.context import FileContent
from scfile.types import OutputDir, PathLike


Content = TypeVar("Content", bound=FileContent)


__all__ = (
    "Content",
    "OutputDir",
    "PathLike",
)



__version__ = "4.2.0"
__author__ = "onejeuu"
__license__ = "MIT"


from . import cli, convert, core, exceptions, formats, structures, consts, enums
from .core.context.options import UserOptions

__all__ = (
    "cli",
    "convert",
    "core",
    "exceptions",
    "formats",
    "structures",
    "consts",
    "enums",
    "UserOptions",
)



"""
Extensions for OL file format with custom struct-based I/O methods.
"""

from scfile.consts import CubemapFaces, OlString
from scfile.core.io import StructFileIO
from scfile.enums import F


class OlFileIO(StructFileIO):
    def _readsizes(self, mipmap_count: int) -> list[int]:
        return [self._readb(F.U32) for _ in range(mipmap_count)]

    def _readsizescubemap(self, mipmap_count: int) -> list[list[int]]:
        return [[self._readb(F.U32) for _ in range(CubemapFaces.COUNT)] for _ in range(mipmap_count)]

    def _readformat(self) -> bytes:
        # Read string and skip last 0x00 byte
        string = self.read(OlString.SIZE)[:-1]

        # Xor byte if it's is not null
        return bytes(byte ^ OlString.XOR for byte in string if byte != OlString.NULL)



"""
Dataclasses for game specific textures.
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import TypeVar

from scfile.consts import CubemapFaces


@dataclass
class Texture(ABC):
    """Base dataclass representing texture."""

    @property
    @abstractmethod
    def image(self) -> bytes:
        pass

    @property
    @abstractmethod
    def linear_size(self) -> int:
        pass


TextureType = TypeVar("TextureType", bound=Texture)


@dataclass
class DefaultTexture(Texture):
    """Standard 2D texture with mipmaps."""

    uncompressed: list[int] = field(default_factory=list)
    compressed: list[int] = field(default_factory=list)
    mipmaps: list[bytes] = field(default_factory=list)

    @property
    def image(self):
        return b"".join(self.mipmaps)

    @property
    def linear_size(self):
        return self.uncompressed[0]


@dataclass
class CubemapTexture(Texture):
    """Cube map texture with face separated mipmaps."""

    uncompressed: list[list[int]] = field(default_factory=list)
    compressed: list[list[int]] = field(default_factory=list)
    faces: list[list[bytes]] = field(default_factory=lambda: [[] for _ in range(CubemapFaces.COUNT)])

    @property
    def image(self):
        return b"".join(b"".join(face) for face in self.faces)

    @property
    def linear_size(self):
        return self.uncompressed[0][0]



"""
Implement functions to convert one format to another by simplest way.
"""

from typing import Optional

from scfile import formats
from scfile.core import UserOptions
from scfile.types import PathLike

from .factory import converter


@converter(formats.mcsb.McsbDecoder, formats.obj.ObjEncoder)
def mcsb_to_obj(
    source: PathLike,
    output: Optional[PathLike] = None,
    options: Optional[UserOptions] = None,
):
    """
    Converts model from `.mcsb` to `.obj` format.

    Arguments:
        source: Path to input `.mcsb` file.
        output (optional): Path to output directory. Defaults: `Same directory as source`.
        options (optional): User settings. Default: `None`.

    Example:
        `mcsb_to_obj("model.mcsb", "path/to/output", UserOptions(parse_skeleton=True))`
    """


@converter(formats.mcsb.McsbDecoder, formats.glb.GlbEncoder)
def mcsb_to_glb(
    source: PathLike,
    output: Optional[PathLike] = None,
    options: Optional[UserOptions] = None,
):
    """
    Converts model from `.mcsb` to `.glb` format.

    Arguments:
        source: Path to input `.mcsb` file.
        output (optional): Path to output directory. Defaults: `Same directory as source`.
        options (optional): User settings. Default: `None`.

    Example:
        `mcsb_to_glb("model.mcsb", "path/to/output", UserOptions(parse_skeleton=True))`
    """


@converter(formats.mcsb.McsbDecoder, formats.dae.DaeEncoder)
def mcsb_to_dae(
    source: PathLike,
    output: Optional[PathLike] = None,
    options: Optional[UserOptions] = None,
):
    """
    Converts model from `.mcsb` to `.dae` format.

    Arguments:
        source: Path to input `.mcsb` file.
        output (optional): Path to output directory. Defaults: `Same directory as source`.
        options (optional): User settings. Default: `None`.

    Example:
        `mcsb_to_dae("model.mcsb", "path/to/output", UserOptions(parse_skeleton=True))`
    """


@converter(formats.mcsb.McsbDecoder, formats.ms3d.Ms3dEncoder)
def mcsb_to_ms3d(
    source: PathLike,
    output: Optional[PathLike] = None,
    options: Optional[UserOptions] = None,
):
    """
    Converts model from `.mcsb` to `.ms3d` format.

    Arguments:
        source: Path to input `.mcsb` file.
        output (optional): Path to output directory. Defaults: `Same directory as source`.
        options (optional): User settings. Default: `None`.

    Example:
        `mcsb_to_ms3d("model.mcsb", "path/to/output", UserOptions(parse_skeleton=True))`
    """


@converter(formats.ol.OlDecoder, formats.dds.DdsEncoder)
def ol_to_dds(
    source: PathLike,
    output: Optional[PathLike] = None,
    options: Optional[UserOptions] = None,
):
    """
    Converts texture from `.ol` to `.dds` format.

    Arguments:
        source: Path to input `.ol` file.
        output (optional): Path to output directory. Defaults: `Same directory as source`.
        options (optional): User settings. Default: `None`.

    Example:
        `ol_to_dds("texture.ol", "path/to/output")`
    """


@converter(formats.hdri.OlCubemapDecoder, formats.dds.DdsEncoder)
def ol_cubemap_to_dds(
    source: PathLike,
    output: Optional[PathLike] = None,
    options: Optional[UserOptions] = None,
):
    """
    Converts cubemap texture from `.ol` to `.dds` format.

    Arguments:
        source: Path to input `.ol` file.
        output (optional): Path to output directory. Defaults: `Same directory as source`.
        options (optional): User settings. Default: `None`.

    Example:
        `ol_cubemap_to_dds("cubemap.ol", "path/to/output")`
    """


@converter(formats.mic.MicDecoder, formats.png.PngEncoder)
def mic_to_png(
    source: PathLike,
    output: Optional[PathLike] = None,
    options: Optional[UserOptions] = None,
):
    """
    Converts image from `.mic` to `.png` format.

    Arguments:
        source: Path to input `.mic` file.
        output (optional): Path to output directory. Defaults: `Same directory as source`.
        options (optional): User settings. Default: `None`.

    Example:
        `mic_to_png("image.mic", "path/to/output")`
    """


@converter(formats.texarr.TextureArrayDecoder, formats.zip.TextureArrayEncoder)
def texarr_to_zip(
    source: PathLike,
    output: Optional[PathLike] = None,
    options: Optional[UserOptions] = None,
):
    """
    Converts archive from `.texarr` to `.zip` format.

    Arguments:
        source: Path to input `.texarr` file.
        output (optional): Path to output directory. Defaults: `Same directory as source`.
        options (optional): User settings. Default: `None`.

    Example:
        `texarr_to_zip("blockMap.texarr", "path/to/output")`
    """


@converter(formats.nbt.NbtDecoder, formats.json.JsonEncoder)
def nbt_to_json(
    source: PathLike,
    output: Optional[PathLike] = None,
    options: Optional[UserOptions] = None,
):
    """
    Converts data from `NBT` to `.json` format.

    Arguments:
        source: Path to input `NBT` file.
        output (optional): Path to output directory. Defaults: `Same directory as source`.
        options (optional): User settings. Default: `None`.

    Example:
        `nbt_to_json("itemnames.dat", "path/to/output")`
    """



"""
Numpy vectors annotation.
"""

from typing import Annotated

import numpy as np
from numpy.typing import NDArray


Vector2D = Annotated[NDArray[np.float32], (..., 2)]
Vector3D = Annotated[NDArray[np.float32], (..., 3)]
Vector4D = Annotated[NDArray[np.float32], (..., 4)]

LinksIds = Annotated[NDArray[np.uint8], (..., 4)]
LinksWeights = Annotated[NDArray[np.float32], (..., 4)]

Polygons = Annotated[NDArray[np.uint32], (..., 3)]



"""
Dataclasses for scene of 3D models.
"""

from collections import defaultdict
from dataclasses import dataclass, field
from typing import TypeAlias

from .animation import ModelAnimation
from .mesh import ModelMesh
from .skeleton import ModelSkeleton


ModelFlags: TypeAlias = defaultdict[int, bool]


@dataclass
class SceneScales:
    """Multiplier values for scene components."""

    position: float = 1.0
    texture: float = 1.0
    unknown: float = 1.0
    filtering: float = 0.1


@dataclass
class SceneCounts:
    """Quantifying of scene elements."""

    meshes: int = 0
    bones: int = 0
    clips: int = 0


@dataclass
class ModelScene:
    """Complete 3D model container with geometry, skeleton and animation."""

    scale: SceneScales = field(default_factory=SceneScales)
    count: SceneCounts = field(default_factory=SceneCounts)

    meshes: list[ModelMesh] = field(default_factory=list)
    skeleton: ModelSkeleton = field(default_factory=ModelSkeleton)
    animation: ModelAnimation = field(default_factory=ModelAnimation)

    @property
    def total_vertices(self):
        return sum(mesh.count.vertices for mesh in self.meshes)

    @property
    def total_polygons(self):
        return sum(mesh.count.polygons for mesh in self.meshes)

    def ensure_unique_names(self):
        seen_names: set[str] = set()

        for mesh in self.meshes:
            name = mesh.name or "noname"

            base_name, count = name, 2
            unique_name = f"{base_name}"

            while unique_name in seen_names:
                unique_name = f"{base_name}_{count}"
                count += 1

            mesh.name = unique_name
            seen_names.add(unique_name)

    def invert_v_textures(self):
        for mesh in self.meshes:
            mesh.textures[:, 1] *= -1

    def flip_v_textures(self):
        for mesh in self.meshes:
            mesh.textures[:, 1] = 1.0 - mesh.textures[:, 1]



from typing import TypeAlias


Version: TypeAlias = float
FlagsCount: TypeAlias = int

VERSION_FLAGS: dict[Version, FlagsCount] = {
    7.0: 4,
    8.0: 5,
    10.0: 6,
    11.0: 6,
}

SUPPORTED_VERSIONS: list[Version] = list(VERSION_FLAGS.keys())



from typing import Callable, ClassVar, Self

from scfile.core.context.content import NbtValue
from scfile.core.io import StructBytesIO
from scfile.enums import ByteOrder, F

from .enums import Tag


class NbtBytesIO(StructBytesIO):
    order: ByteOrder = ByteOrder.BIG

    _HANDLERS: ClassVar[dict[Tag, Callable[[Self], NbtValue]]] = {
        Tag.END: lambda _: None,
        Tag.BYTE: lambda s: s._readb(F.I8),
        Tag.SHORT: lambda s: s._readb(F.I16),
        Tag.INT: lambda s: s._readb(F.I32),
        Tag.LONG: lambda s: s._readb(F.I64),
        Tag.FLOAT: lambda s: s._readb(F.F32),
        Tag.DOUBLE: lambda s: s._readb(F.F64),
        Tag.BYTE_ARRAY: lambda s: s._read_byte_array(),
        Tag.STRING: lambda s: s._readutf8(),
        Tag.LIST: lambda s: s._read_list(),
        Tag.COMPOUND: lambda s: s._read_compound(),
        Tag.INT_ARRAY: lambda s: s._read_int_array(),
        Tag.LONG_ARRAY: lambda s: s._read_long_array(),
    }

    def _parse_tag(self, tag: Tag) -> NbtValue:
        return self._HANDLERS[tag](self)

    def _read_tag(self) -> Tag:
        return Tag(self._readb(F.I8))

    def _read_byte_array(self) -> bytes:
        length = self._readb(F.I32)
        return self.read(length)

    def _read_list(self) -> list[NbtValue]:
        tag = self._read_tag()
        length = self._readb(F.I32)
        return [self._parse_tag(tag) for _ in range(length)]

    def _read_int_array(self) -> list[int]:
        length = self._readb(F.I32)
        return [self._readb(F.I32) for _ in range(length)]

    def _read_long_array(self) -> list[int]:
        length = self._readb(F.I32)
        return [self._readb(F.I64) for _ in range(length)]

    def _read_compound(self) -> dict[str, NbtValue]:
        data = {}
        while True:
            tag = self._read_tag()
            if tag == Tag.END:
                break
            key = self._readutf8()
            data[key] = self._parse_tag(tag)
        return data



"""
Dataclasses for 3D model armature.
"""

from dataclasses import dataclass, field
from typing import List, Self

import numpy as np

from scfile.consts import McsaModel

from .vectors import Vector3D, Vector4D


@dataclass
class SkeletonBone:
    """Single skeleton bone with transform data."""

    id: int = 0
    name: str = "bone"
    parent_id: int = McsaModel.ROOT_BONE_ID

    position: Vector3D = field(default_factory=lambda: np.empty(3, dtype=np.float32))
    rotation: Vector3D = field(default_factory=lambda: np.empty(3, dtype=np.float32))

    children: List[Self] = field(default_factory=list, repr=False)

    @property
    def is_root(self) -> bool:
        return self.parent_id == McsaModel.ROOT_BONE_ID

    @property
    def quaternion(self) -> Vector4D:
        return euler_to_quat(self.rotation)


@dataclass
class ModelSkeleton:
    """Skeleton bones container."""

    bones: List[SkeletonBone] = field(default_factory=list)

    @property
    def roots(self) -> List[SkeletonBone]:
        return list(filter(lambda bone: bone.is_root, self.bones))

    def convert_to_local(self) -> None:
        for bone in self.bones:
            parent_id = bone.parent_id

            # Convert global position to parent-relative space
            while parent_id > McsaModel.ROOT_BONE_ID:
                parent = self.bones[parent_id]
                bone.position -= parent.position  # Make position relative to parent
                parent_id = parent.parent_id  # Move up hierarchy

    def build_hierarchy(self) -> None:
        # Rebuild hierarchy
        for bone in self.bones:
            if not bone.is_root:
                parent = self.bones[bone.parent_id]
                parent.children.append(bone)

    def calculate_global_transforms(self) -> list[np.ndarray]:
        transforms: list[np.ndarray] = []

        for bone in self.bones:
            local_matrix = create_transform_matrix(bone.position, bone.rotation)
            global_matrix = local_matrix if bone.is_root else transforms[bone.parent_id] @ local_matrix
            transforms.append(global_matrix)

        return transforms

    def inverse_bind_matrices(self, transpose: bool) -> np.ndarray:
        global_transforms = self.calculate_global_transforms()

        inverse_matrices = [np.linalg.inv(matrix) for matrix in global_transforms]

        if transpose:
            inverse_matrices = [matrix.T for matrix in inverse_matrices]

        return np.array(inverse_matrices, dtype=np.float32)


def create_rotation_matrix(rotation: Vector3D) -> np.ndarray:
    angles = np.radians(rotation)
    cx, cy, cz = np.cos(angles)
    sx, sy, sz = np.sin(angles)

    return np.array(
        [
            [cy * cz, -cy * sz, sy],
            [cx * sz + cz * sx * sy, cx * cz - sx * sy * sz, -cy * sx],
            [sx * sz - cx * cz * sy, cz * sx + cx * sy * sz, cx * cy],
        ],
        dtype=np.float32,
    )


def create_transform_matrix(position: Vector3D, rotation: Vector3D) -> np.ndarray:
    matrix = np.eye(4, dtype=np.float32)
    matrix[:3, :3] = create_rotation_matrix(rotation)
    matrix[:3, 3] = position
    return matrix


def euler_to_quat(rotation: Vector3D, degrees: bool = True) -> Vector4D:
    x, y, z = np.radians(rotation) if degrees else rotation
    hx, hy, hz = x * 0.5, y * 0.5, z * 0.5

    cx, cy, cz = np.cos([hx, hy, hz])
    sx, sy, sz = np.sin([hx, hy, hz])

    return np.array(
        [
            sx * cy * cz - cx * sy * sz,
            cx * sy * cz + sx * cy * sz,
            cx * cy * sz - sx * sy * cz,
            cx * cy * cz + sx * sy * sz,
        ],
        dtype=np.float32,
    )



from enum import IntFlag


class HeaderFlag(IntFlag):
    CAPS = 0x1
    HEIGHT = 0x2
    WIDTH = 0x4
    PITCH = 0x8
    PIXELFORMAT = 0x1000
    MIPMAPCOUNT = 0x20000
    LINEARSIZE = 0x80000
    DEPTH = 0x800000


class PixelFormatFlag(IntFlag):
    ALPHAPIXELS = 0x1
    ALPHA = 0x2
    FOURCC = 0x4
    RGB = 0x40


class CubemapFlag(IntFlag):
    POSITIVE_X = POSX = 0x400
    NEGATIVE_X = NEGX = 0x800
    POSITIVE_Y = POSY = 0x1000
    NEGATIVE_Y = NEGY = 0x2000
    POSITIVE_Z = POSZ = 0x4000
    NEGATIVE_Z = NEGZ = 0x8000


class DXGIFormat(IntFlag):
    FLOAT_R32G32B32A32 = 2
    FLOAT_R32G32B32 = 6


class DXGIDimension(IntFlag):
    BUFFER = 1
    TEXTURE1D = 2
    TEXTURE2D = 3
    TEXTURE3D = 4



"""
Implement functions to convert legacy mcsa format to another by simplest way.
"""

from typing import Optional

from scfile import formats
from scfile.core import UserOptions
from scfile.types import PathLike

from .formats import converter


@converter(formats.mcsa.McsaDecoder, formats.obj.ObjEncoder)
def mcsa_to_obj(
    source: PathLike,
    output: Optional[PathLike] = None,
    options: Optional[UserOptions] = None,
):
    """
    Converts model from `.mcsa` to `.obj` format.

    Arguments:
        source: Path to input `.mcsa` file.
        output (optional): Path to output directory. Defaults: `Same directory as source`.
        options (optional): User settings. Default: `None`.

    Example:
        `mcsa_to_obj("model.mcsa", "path/to/output", UserOptions(parse_skeleton=True))`
    """


@converter(formats.mcsa.McsaDecoder, formats.glb.GlbEncoder)
def mcsa_to_glb(
    source: PathLike,
    output: Optional[PathLike] = None,
    options: Optional[UserOptions] = None,
):
    """
    Converts model from `.mcsa` to `.glb` format.

    Arguments:
        source: Path to input `.mcsa` file.
        output (optional): Path to output directory. Defaults: `Same directory as source`.
        options (optional): User settings. Default: `None`.

    Example:
        `mcsa_to_glb("model.mcsa", "path/to/output", UserOptions(parse_skeleton=True))`
    """


@converter(formats.mcsa.McsaDecoder, formats.dae.DaeEncoder)
def mcsa_to_dae(
    source: PathLike,
    output: Optional[PathLike] = None,
    options: Optional[UserOptions] = None,
):
    """
    Converts model from `.mcsa` to `.dae` format.

    Arguments:
        source: Path to input `.mcsa` file.
        output (optional): Path to output directory. Defaults: `Same directory as source`.
        options (optional): User settings. Default: `None`.

    Example:
        `mcsa_to_dae("model.mcsa", "path/to/output", UserOptions(parse_skeleton=True))`
    """


@converter(formats.mcsa.McsaDecoder, formats.ms3d.Ms3dEncoder)
def mcsa_to_ms3d(
    source: PathLike,
    output: Optional[PathLike] = None,
    options: Optional[UserOptions] = None,
):
    """
    Converts model from `.mcsa` to `.ms3d` format.

    Arguments:
        source: Path to input `.mcsa` file.
        output (optional): Path to output directory. Defaults: `Same directory as source`.
        options (optional): User settings. Default: `None`.

    Example:
        `mcsa_to_ms3d("model.mcsa", "path/to/output", UserOptions(parse_skeleton=True))`
    """



"""
Base class for file encoder (serialization).
"""

from abc import ABC, abstractmethod
from typing import Any, Generic, Optional, Self, TypeAlias

from scfile.enums import FileMode
from scfile.structures.flags import Flag
from scfile.types import PathLike

from .base import BaseFile
from .context import ModelContent, UserOptions
from .io import StructBytesIO
from .types import Content


EncoderContext: TypeAlias = dict[str, Any]


class FileEncoder(BaseFile, StructBytesIO, Generic[Content], ABC):
    """Base class for encoding structured data objects into file content."""

    @property
    def mode(self) -> str:
        return FileMode.WRITE

    def __init__(self, data: Content, options: Optional[UserOptions] = None):
        """Initialize file encoder with content data and options.

        Arguments:
            data: Content data to be encoded.
            options (optional): User provided options. If None, default `UserOptions` will be used.

        Initialized:
            data (`Generic[Content]`): Content to encode.
            options (`UserOptions`): Encoding options (default or user provided).
            ctx (`EncoderContext`): Empty context dictionary for processing state.

        Note:
            Actual encoding doesn't happen during initialization.
            Call `encode()` to perform serialization process.
        """

        self.data: Content = data
        self.options: UserOptions = options or UserOptions()
        self.ctx: EncoderContext = {}

    @property
    def suffix(self) -> str:
        """Return standard file extension for this format (with dot)."""
        return self.format.suffix

    def encode(self) -> Self:
        """Encode data: prepare, add signature, serialize. Returns self."""
        self.prepare()
        self.add_signature()
        self.serialize()
        return self

    def encoded(self) -> Self:
        """Context manager that automatically encodes data on exit."""
        return self.encode()

    def prepare(self) -> None:
        """Perform preparations before serialization. *(e.g. calculations in content)*."""
        pass

    def add_signature(self) -> None:
        """Write format signature (magic bytes) to buffer if defined."""
        if self.signature:
            self.write(self.signature)

    @abstractmethod
    def serialize(self) -> None:
        """Convert structured `self.data` into bytes and write to buffer."""
        pass

    def save_as(self, path: PathLike) -> None:
        """Write buffer content to file. Keeps encoder open."""
        with open(path, mode=self.mode) as fp:
            fp.write(self.getvalue())

    def export_as(self, path: PathLike) -> None:
        """Save to path (without suffix) adding format suffix automatically. Keeps encoder open."""
        self.save_as(path=f"{path}{self.suffix}")

    def save(self, path: PathLike) -> None:
        """Write buffer content to file. Closes encoder."""
        self.save_as(path=path)
        self.close()

    def export(self, path: PathLike) -> None:
        """Save to path (without suffix) adding format suffix automatically. Closes encoder."""
        self.save(path=f"{path}{self.suffix}")
        self.close()

    def close(self) -> None:
        """Close buffer and reset `self.ctx` state. Same as `BytesIO.close()`."""
        self.ctx = {}
        super().close()

    @property
    def _skeleton_presented(self) -> bool:
        if isinstance(self.data, ModelContent):
            return self.data.flags[Flag.SKELETON] and self.options.parse_skeleton
        return False

    @property
    def _animation_presented(self) -> bool:
        return self._skeleton_presented and self.options.parse_animation
